/****************************************************************************
*                tokenize.c
*
*  This module implements the first part of a two part parser for the scene
*  description files.  This phase changes the input file into tokens.
*
*  This module tokenizes the input file and sends the tokens created
*  to the parser (the second stage).  Tokens sent to the parser contain a
*  token ID, the line number of the token, and if necessary, some data for
*  the token.
*
*  from Persistence of Vision(tm) Ray Tracer
*  Copyright 1996,1999 Persistence of Vision Team
*---------------------------------------------------------------------------
*  NOTICE: This source code file is provided so that users may experiment
*  with enhancements to POV-Ray and to port the software to platforms other
*  than those supported by the POV-Ray Team.  There are strict rules under
*  which you are permitted to use this file.  The rules are in the file
*  named POVLEGAL.DOC which should be distributed with this file.
*  If POVLEGAL.DOC is not available or for more info please contact the POV-Ray
*  Team Coordinator by email to team-coord@povray.org or visit us on the web at
*  http://www.povray.org. The latest version of POV-Ray may be found at this site.
*
* This program is based on the popular DKB raytracer version 2.12.
* DKBTrace was originally written by David K. Buck.
* DKBTrace Ver 2.0-2.12 were written by David K. Buck & Aaron A. Collins.
*
* Modified by Edward Coffey, September 1999 - fade_colour added
*
*  WARNING: This is a changed copy of the original file.
*  Support of sphere sweep primitive added by Jochen Lippert.
*
*****************************************************************************/

#include <ctype.h>
#include "frame.h"
#ifdef FastMacroPatch
#include <stdarg.h> /* Portable?  (Burton Radons, Jan. 2000) */
#endif
#include "povray.h"
#include "povproto.h"
#include "parse.h"
#include "povray.h"
#include "colour.h"
#include "render.h"
#include "texture.h"
#include "tokenize.h"
#include "express.h"
#include "matrices.h"
#ifdef UnofficialBlocking
#include "optout.h"
#endif

/** poviso: Mar 99 R.S. **/
#ifdef POVISO
extern int Func_Parse_Flag;
#endif
/** --- **/
#ifdef CreateLabelPatch
/* NK persist create Nov 1999 */
extern char *createObjectName;
/* NK ---- */
#endif


/*****************************************************************************
* Local preprocessor defines
******************************************************************************/



/*****************************************************************************
* Local typedefs
******************************************************************************/

#define MAX_INCLUDE_FILES 32

#define CALL(x) { if (!(x)) return (FALSE); }

#define COND_STACK_SIZE 200

typedef enum cond_type
{
  ROOT_COND=0,
  WHILE_COND,
  IF_TRUE_COND,
  IF_FALSE_COND,
  ELSE_COND,
  SWITCH_COND,
  CASE_TRUE_COND,
  CASE_FALSE_COND,
  SKIP_TIL_END_COND,
  INVOKING_MACRO_COND,
  DECLARING_MACRO_COND
} COND_TYPE;


#define SYM_TABLE_SIZE 257
#define MAX_NUMBER_OF_TABLES 100


/*****************************************************************************
* Local variables
******************************************************************************/

static DATA_FILE *Temp_Data_File_For_Fread;

typedef struct Sym_Table_Struct SYM_TABLE;

struct Sym_Table_Struct
{
  char *Table_Name;
  SYM_ENTRY *Table[SYM_TABLE_SIZE];
};

SYM_TABLE *Tables[MAX_NUMBER_OF_TABLES];

int Table_Index;

static int String_Index;
char String[MAX_STRING_INDEX];
char String2[MAX_STRING_INDEX];

/* moved here to allow reinitialization */

int token_count = 0;
int line_count = 10;

static int Include_File_Index;
/* static   - 0. */
DATA_FILE *Data_File;
/* 0 */
static DATA_FILE Include_Files[MAX_INCLUDE_FILES];

struct Token_Struct Token;

static char **Echo_Buff;
static char  *Echo_Ptr;
static int    Echo_Indx;
static int    Echo_Line;

#define MAX_PARAMETER_LIST 20

typedef struct Cond_Stack_Entry CS_ENTRY;

struct Cond_Stack_Entry
{
  COND_TYPE Cond_Type;
  DBL Switch_Value;
  char *While_File_Name;
  char *Macro_Return_Name;
  int Macro_Same_Flag;
  POV_MACRO *PMac;
  long Pos,Line_No;
#ifdef UnofficialBlocking
  int unofficialVersion;
#endif
#ifdef FastMacroPatch
  DATA_FILE Return;
#endif
};

static CS_ENTRY *Cond_Stack;
static int Skipping, Got_EOF, Inside_Ifdef, Inside_MacroDef;
int CS_Index;

int input_file_in_memory = 0 ;

/*
 * Here are the reserved words.  If you need to add new words,
 * be sure to declare them in parse.h
 */

RESERVED_WORD Reserved_Words [LAST_TOKEN] = {
  {ABSORPTION_TOKEN, "absorption"},
  {ABS_TOKEN, "abs"},
  {ACOSH_TOKEN, "acosh"},
  {ACOS_TOKEN, "acos"},
  {ADAPTIVE_TOKEN, "adaptive"},
  {ADC_BAILOUT_TOKEN, "adc_bailout"},
  {AGATE_TOKEN, "agate"},
  {AGATE_TURB_TOKEN, "agate_turb"},
  {ALL_TOKEN, "all"},
  {ALPHA_TOKEN, "alpha"},
  {AMBIENT_LIGHT_TOKEN, "ambient_light"},
  {AMBIENT_TOKEN, "ambient"},
  {AMPERSAND_TOKEN, "&"},
  {ANGLE_TOKEN, "angle"},
  {APERTURE_TOKEN, "aperture"},
  {APPEND_TOKEN, "append"},
  {ARC_ANGLE_TOKEN, "arc_angle"},
  {AREA_LIGHT_TOKEN, "area_light"},
#ifdef CircularOrientAreaLightPatch 
  {AREA_LIGHT_CIRCULAR_TOKEN,"circular"}, /* Circular area lights [ENB 9/97] */
  {AREA_LIGHT_ORIENT_TOKEN,"orient"},     /* Orient area lights   [ENB 9/97] */
#endif
  {ARRAY_ID_TOKEN, "array identifier"},
  {ARRAY_TOKEN, "array"},
  {ASC_TOKEN, "asc"},
  {ASINH_TOKEN, "asinh"},
  {ASIN_TOKEN, "asin"},
  {ASSUMED_GAMMA_TOKEN, "assumed_gamma"},
  {ATAN2_TOKEN, "atan2"},
  {ATANH_TOKEN, "atanh"},
  {ATAN_TOKEN, "atan"},
#ifdef AtanWavePatch
  {ATAN_WAVE_TOKEN, "atan_wave"},
#endif  
  {AT_TOKEN, "@"},
  {AVERAGE_TOKEN, "average"},
  {BACKGROUND_TOKEN, "background"},
  {BACK_QUOTE_TOKEN, "`"},
  {BACK_SLASH_TOKEN, "\\"},
  {BAR_TOKEN, "|"},
#ifdef RBezierPatch

  {BEZIER_PATCH_TOKEN, "bezier_patch"},
#endif
  {BEZIER_SPLINE_TOKEN, "bezier_spline"},
  {BICUBIC_PATCH_TOKEN, "bicubic_patch"},
  {BLACK_HOLE_TOKEN, "black_hole"},
  {BLOB_TOKEN, "blob"},
  {BLUE_TOKEN, "blue"},
  {BLUR_SAMPLES_TOKEN, "blur_samples"},
  {BOUNDED_BY_TOKEN, "bounded_by"},
  {BOXED_TOKEN, "boxed"},
  {BOX_TOKEN, "box"},
  {BOZO_TOKEN, "bozo"},
  {BREAK_TOKEN, "break"},
  {BRICK_SIZE_TOKEN, "brick_size"},
  {BRICK_TOKEN, "brick"},
  {BRIGHTNESS_TOKEN, "brightness" },
  {BRILLIANCE_TOKEN, "brilliance"},
  {BUMPS_TOKEN, "bumps"},
  {BUMP_MAP_TOKEN, "bump_map"},
  {BUMP_SIZE_TOKEN, "bump_size"},

  {CAMERA_ID_TOKEN, "camera identifier"},
  {CAMERA_TOKEN, "camera"},
#ifdef PostCartoonPatch
  {CARTOON_TOKEN, "cartoon"},
#endif
  {CASE_TOKEN, "case"},
  {CAUSTICS_TOKEN, "caustics"},
  {CEIL_TOKEN, "ceil"},
  {CHECKER_TOKEN, "checker"},
  {CHR_TOKEN, "chr"},
  {CLIPPED_BY_TOKEN, "clipped_by"},
  {CLOCK_DELTA_TOKEN, "clock_delta"},
  {CLOCK_TOKEN, "clock"},
 #ifdef PostProcessCurvesPatch
  {CURVES_TOKEN, "curves"},
#endif  

#ifdef HfHeightAtPatch
  /* ADDED 2/18/98 by JB for height_field functions */
  {HF_HEIGHT_AT_TOKEN, "hf_height_at"},
#endif
#ifdef GetClockValuesPatch
  {CLOCK_ON_TOKEN, "clock_on"},
  {FRAMENUMBER_TOKEN, "frame_number"},
  {INITIALFRAME_TOKEN, "initial_frame"},
  {FINALFRAME_TOKEN, "final_frame"},
  {INITIALCLOCK_TOKEN, "initial_clock"},
  {FINALCLOCK_TOKEN, "final_clock"},
#endif
#ifdef GetTicksPatch
  {START_CHRONO_TOKEN,"start_chrono"},
  {CURRENT_CHRONO_TOKEN, "current_chrono"},
  {GET_CLK_TCK_TOKEN, "tick_count"},
#endif
 #ifdef TimeDatePatch
  {GET_DATE_TOKEN, "date"}, /*YS get date and time string*/
#endif

#ifdef GetImageSizePatch
  {IMAGE_WIDTH_TOKEN,"image_width"},
  {IMAGE_HEIGHT_TOKEN,"image_height"},
#endif
#ifdef UnofficialBlocking
  {UNOFFICIAL_VERSION_TOKEN,"unofficial_version"},
#endif


  {COLON_TOKEN, ":"},
  {COLOUR_ID_TOKEN, "colour identifier"},
  {COLOUR_KEY_TOKEN, "color keyword"},
  {COLOUR_MAP_ID_TOKEN, "colour map identifier"},
  {COLOUR_MAP_TOKEN, "color_map"},
  {COLOUR_MAP_TOKEN, "colour_map"},
  {COLOUR_TOKEN, "color"},
  {COLOUR_TOKEN, "colour"},
  {COMMA_TOKEN, ", "},
  {COMPONENT_TOKEN, "component"},
  {COMPOSITE_TOKEN, "composite"},
  {CONCAT_TOKEN, "concat"},
  {CONE_TOKEN, "cone"},
  {CONFIDENCE_TOKEN, "confidence"},
  {CONIC_SWEEP_TOKEN, "conic_sweep"},
  {CONTROL0_TOKEN, "control0"},
  {CONTROL1_TOKEN, "control1"},
  {COSH_TOKEN, "cosh"},
  {COS_TOKEN, "cos"},
  {COUNT_TOKEN, "count" },
  {CRACKLE_TOKEN, "crackle"},
  {CRAND_TOKEN, "crand"},
  {CUBE_TOKEN, "cube"},
  {CUBIC_SPLINE_TOKEN, "cubic_spline"},
  {CUBIC_TOKEN, "cubic"},
  {CUBIC_WAVE_TOKEN, "cubic_wave"},
  {CYLINDER_TOKEN, "cylinder"},
  {CYLINDRICAL_TOKEN, "cylindrical"},
  {DASH_TOKEN, "-"},
  {DEBUG_TOKEN, "debug"},
  {DECLARE_TOKEN, "declare"},
  {DEFAULT_TOKEN, "default"},
  {DEFINED_TOKEN, "defined"},
  {DEGREES_TOKEN, "degrees"},
  {DENSITY_FILE_TOKEN, "density_file"},
  {DENSITY_ID_TOKEN, "density identifier"},
  {DENSITY_MAP_ID_TOKEN, "density_map identifier"},
  {DENSITY_MAP_TOKEN, "density_map"},
  {DENSITY_TOKEN, "density"},
  {DENTS_TOKEN, "dents"},
  {DF3_TOKEN, "df3"},
  {DIFFERENCE_TOKEN, "difference"},
  {DIFFUSE_TOKEN, "diffuse"},
  {DIMENSIONS_TOKEN, "dimensions"},
  {DIMENSION_SIZE_TOKEN, "dimension_size"},
  {DIRECTION_TOKEN, "direction"},
  {DISC_TOKEN, "disc"},
  {DISTANCE_MAXIMUM_TOKEN, "distance_maximum" },
  {DISTANCE_TOKEN, "distance"},
  {DIV_TOKEN, "div"},
  {DOLLAR_TOKEN, "$"},
  {ECCENTRICITY_TOKEN, "eccentricity"},
#ifdef BsplinePatch
  {EVAL_3D_SPLINE_TOKEN,"eval_3d_spline"},
  {EVAL_SPLINE_TOKEN,"eval_spline"},
#endif
  {ELSE_TOKEN, "else"},
  {EMISSION_TOKEN, "emission"},
  {EMPTY_ARRAY_TOKEN, "empty array"},
  {END_OF_FILE_TOKEN, "End of File"},
  {END_TOKEN, "end"},
  {EQUALS_TOKEN, "="},
  {ERROR_BOUND_TOKEN, "error_bound" },
  {ERROR_TOKEN, "error"},
  {EXCLAMATION_TOKEN, "!"},
  {EXP_TOKEN, "exp"},
  {EXTINCTION_TOKEN, "extinction"},
  {FADE_DISTANCE_TOKEN, "fade_distance"},
  {FADE_POWER_TOKEN, "fade_power"},
  {FADE_COLOUR_TOKEN, "fade_colour"},
  {FADE_COLOUR_TOKEN, "fade_color"},
  {FALLOFF_ANGLE_TOKEN, "falloff_angle"},
  {FALLOFF_TOKEN, "falloff"},
  {FALSE_TOKEN, "false"},
  {FCLOSE_TOKEN, "fclose"},
  {FILE_EXISTS_TOKEN, "file_exists"},
  {FILE_ID_TOKEN, "file identifier"},
  {FILL_LIGHT_TOKEN, "shadowless"},
  {FILTER_TOKEN, "filter"},
  {FINISH_ID_TOKEN, "finish identifier"},
  {FINISH_TOKEN, "finish"},
#ifdef RBezierPatch

  {FIRST_TOKEN, "first"},
#endif
  {FISHEYE_TOKEN, "fisheye"},
  {FLATNESS_TOKEN, "flatness"},
  {FLIP_TOKEN, "flip"},
  {FLOAT_FUNCT_TOKEN, "float function"},
  {FLOAT_ID_TOKEN, "float identifier"},
  {FLOAT_TOKEN, "float constant"},
  {FLOOR_TOKEN, "floor"},
  {FOCAL_POINT_TOKEN, "focal_point"},
  {FOG_ALT_TOKEN, "fog_alt"},
  {FOG_ID_TOKEN, "fog identifier"},
  {FOG_OFFSET_TOKEN, "fog_offset"},
  {FOG_TOKEN, "fog"},
  {FOG_TYPE_TOKEN, "fog_type"},
  {FOPEN_TOKEN, "fopen"},
  {FREQUENCY_TOKEN, "frequency"},
#ifdef FontAnlignmentPatch
  {POSITION_TOKEN, "position"},
  {ALIGN_LEFT_TOKEN, "align_left"},
  {ALIGN_RIGHT_TOKEN, "align_right"},
  {ALIGN_CENTER_TOKEN, "align_center"},
#endif
  {GIF_TOKEN, "gif"},
  {GLOBAL_SETTINGS_TOKEN, "global_settings" },
  {GRADIENT_TOKEN, "gradient"},
  {GRANITE_TOKEN, "granite"},
  {GRAY_THRESHOLD_TOKEN, "gray_threshold" },
  {GREEN_TOKEN, "green"},
  {H_ANGLE_TOKEN, "h_angle"},
  {HASH_TOKEN, "#"},
  {HAT_TOKEN, "^"},
  {HEIGHT_FIELD_TOKEN, "height_field"},
  {HEXAGON_TOKEN, "hexagon"},
  {HF_GRAY_16_TOKEN, "hf_gray_16" },
  {HIERARCHY_TOKEN, "hierarchy"},
  {HOLLOW_TOKEN, "hollow"},
  {HYPERCOMPLEX_TOKEN, "hypercomplex"},
  {IDENTIFIER_TOKEN, "undeclared identifier"},
  {IFDEF_TOKEN, "ifdef"},
  {IFF_TOKEN, "iff"},
  {IFNDEF_TOKEN, "ifndef"},
  {IF_TOKEN, "if"},
  {IMAGE_MAP_TOKEN, "image_map"},
  {INCLUDE_TOKEN, "include"},
 #ifdef BsplinePatch
    {INIT_3D_SPLINE_TOKEN, "init_3d_spline"},
    {INIT_SPLINE_TOKEN,"init_spline"},  
#endif
  {INTERIOR_ID_TOKEN, "interior identifier"},
  {INTERIOR_TOKEN, "interior"},
  {INTERPOLATE_TOKEN, "interpolate"},
  {INTERSECTION_TOKEN, "intersection"},
  {INTERVALS_TOKEN, "intervals"},
  {INT_TOKEN, "int"},
  {INVERSE_TOKEN, "inverse"},
  {IOR_TOKEN, "ior"},
  {IRID_TOKEN, "irid"},
  {IRID_WAVELENGTH_TOKEN, "irid_wavelength"},
  {JITTER_TOKEN, "jitter"},
  {JULIA_FRACTAL_TOKEN, "julia_fractal"},

  {LAMBDA_TOKEN, "lambda"},
  {LATHE_TOKEN, "lathe"},
  {LEFT_ANGLE_TOKEN, "<"},
  {LEFT_CURLY_TOKEN, "{"},
  {LEFT_PAREN_TOKEN, "("},
  {LEFT_SQUARE_TOKEN, "["},
  {LEOPARD_TOKEN, "leopard"},
  {LIGHT_SOURCE_TOKEN, "light_source"},
#ifdef PostProcessLimitBrightnessPatch
  {LIMIT_BRIGHTNESS_TOKEN, "limit_brightness"},
#endif
  {LINEAR_SPLINE_TOKEN, "linear_spline"},
  {LINEAR_SWEEP_TOKEN, "linear_sweep"},
  {LOCAL_TOKEN, "local"},
  {LOCATION_TOKEN, "location"},
  {LOG_TOKEN, "log"},
  {LOOKS_LIKE_TOKEN, "looks_like"},
  {LOOK_AT_TOKEN, "look_at"},
  {LOW_ERROR_FACTOR_TOKEN, "low_error_factor" },
  {MACRO_ID_TOKEN, "macro identifier"},
  {MACRO_TOKEN, "macro"},
  {MANDEL_TOKEN, "mandel"},
  {MAP_TYPE_TOKEN, "map_type"},
  {MARBLE_TOKEN, "marble"},
  {MATERIAL_ID_TOKEN, "material identifier"},
  {MATERIAL_MAP_TOKEN, "material_map"},
  {MATERIAL_TOKEN, "material"},
  {MATRIX_TOKEN, "matrix"},
  #ifdef BoundPatch
    {MAX_EXTENT_TOKEN, "max_extent" },
#endif
  {MAX_INTERSECTIONS, "max_intersections"},
  {MAX_ITERATION_TOKEN, "max_iteration"},
  {MAX_TOKEN, "max"},
  {MAX_TRACE_LEVEL_TOKEN, "max_trace_level"},
  {MEDIA_ATTENUATION_TOKEN, "media_attenuation"},
  {MEDIA_ID_TOKEN, "media identifier"},
  {MEDIA_INTERACTION_TOKEN, "media_interaction"},
  {MEDIA_TOKEN, "media"},
  {MERGE_TOKEN, "merge"},
  {MESH_TOKEN, "mesh"},
  {METALLIC_TOKEN, "metallic"},
  {MINIMUM_REUSE_TOKEN, "minimum_reuse" },
  #ifdef BoundPatch
    {MIN_EXTENT_TOKEN, "min_extent" },
#endif
  {MIN_TOKEN, "min"},
  {MOD_TOKEN, "mod"},
  {MORTAR_TOKEN, "mortar"},
  {NEAREST_COUNT_TOKEN, "nearest_count" },
  {NORMAL_MAP_ID_TOKEN, "normal_map identifier"},
  {NORMAL_MAP_TOKEN, "normal_map"},
  {NO_SHADOW_TOKEN, "no_shadow"},
#ifdef NoImageNoReflectionPatch
    {NO_IMAGE_TOKEN, "no_image"},           /* Object-Ray Options [ENB 9/97] */
    {NO_REFLECTION_TOKEN, "no_reflection"}, /* Object-Ray Options [ENB 9/97] */
#endif
  {NO_TOKEN, "no"},
  {NUMBER_OF_WAVES_TOKEN, "number_of_waves"},
  {OBJECT_ID_TOKEN, "object identifier"},
  {OBJECT_TOKEN, "object"},
  {OCTAVES_TOKEN, "octaves"},
  {OFFSET_TOKEN, "offset"},
  {OFF_TOKEN, "off"},
  {OMEGA_TOKEN, "omega"},
  {OMNIMAX_TOKEN, "omnimax"},
  {ONCE_TOKEN, "once"},
  {ONION_TOKEN, "onion"},
  {ON_TOKEN, "on"},
  {OPEN_TOKEN, "open"},
  {ORTHOGRAPHIC_TOKEN, "orthographic"},
  {PANORAMIC_TOKEN, "panoramic"},
  {PARAMETER_ID_TOKEN, "parameter identifier"},
  {PERCENT_TOKEN, "%"},
  {PERIOD_TOKEN, ". (period)"},
  {PERSPECTIVE_TOKEN, "perspective"},
  {PGM_TOKEN, "pgm"},
  {PHASE_TOKEN, "phase"},
  {PHONG_SIZE_TOKEN, "phong_size"},
  {PHONG_TOKEN, "phong"},
  {PIGMENT_ID_TOKEN, "pigment identifier"},
  {PIGMENT_MAP_ID_TOKEN, "pigment_map identifier"},
  {PIGMENT_MAP_TOKEN, "pigment_map"},
  {PIGMENT_TOKEN, "pigment"},
  {PI_TOKEN, "pi"},
#ifdef EvalPatternPatch
  {EVALPAT_TOKEN, "eval_pattern"}, /*Chris Huff eval_pattern function*/
#endif
#ifdef EvalPigmentPatch
  {EVALPIG_TOKEN, "eval_pigment"}, /*Chris Huff eval_pigment function*/
#endif
  {PLANAR_TOKEN, "planar"},
  {PLANE_TOKEN, "plane"},
  {PLUS_TOKEN, "+"},
  {PNG_TOKEN, "png"},
  {POINT_AT_TOKEN, "point_at"},
#ifdef PolaricalPatch
    {POLARICAL_TOKEN, "polarical"}, /* Added by bajcik */
#endif
  {POLYGON_TOKEN, "polygon"},
  {POLY_TOKEN, "poly"},
  {POLY_WAVE_TOKEN, "poly_wave"},
  {POT_TOKEN, "pot"},
  {POW_TOKEN, "pow"},
  {PPM_TOKEN, "ppm"},
  {PRECISION_TOKEN, "precision"},
#ifdef RBezierPatch
  {PREVIOUS_TOKEN,"previous"},
#endif
#ifdef VtransformPatch
  {VTRANSFORM_TOKEN, "vtransform"},
  {VINV_TRANSFORM_TOKEN, "vinv_transform"},
#endif
#ifdef VWarpPatch
  {VWARP_TOKEN, "vwarp"},
#endif
  {PRISM_TOKEN, "prism"},
  {PWR_TOKEN, "pwr"},
  {QUADRATIC_SPLINE_TOKEN, "quadratic_spline"},
  {QUADRIC_TOKEN, "quadric"},
  {QUARTIC_TOKEN, "quartic"},
  {QUATERNION_TOKEN, "quaternion"},
  {QUESTION_TOKEN, "?"},
  {QUICK_COLOUR_TOKEN, "quick_color"},
  {QUICK_COLOUR_TOKEN, "quick_colour"},
  {QUILTED_TOKEN, "quilted"},
  {RADIAL_TOKEN, "radial"},
  {RADIANS_TOKEN, "radians"},
  {RADIOSITY_TOKEN, "radiosity" },
  {RADIUS_TOKEN, "radius"},
  {RAINBOW_ID_TOKEN, "rainbow identifier"},
  {RAINBOW_TOKEN, "rainbow"},
  {RAMP_WAVE_TOKEN, "ramp_wave"},
  {RAND_TOKEN, "rand"},
  {RANGE_TOKEN, "range"},
#ifdef RBezierPatch
  {RATIONAL_TOKEN,"rational"},
#endif
  {RATIO_TOKEN, "ratio"},
#ifdef PostProcessRawImagePatch
  {RAW_IMAGE_TOKEN, "raw_image"},
#endif
  {READ_TOKEN, "read"},
  {RECIPROCAL_TOKEN, "reciprocal" },
  {RECURSION_LIMIT_TOKEN, "recursion_limit" },
  {RED_TOKEN, "red"},
  {REFLECTION_TOKEN, "reflection"},
  {REFLECTION_EXPONENT_TOKEN, "reflection_exponent"},
  {REFRACTION_TOKEN, "refraction"},
  {REL_GE_TOKEN, ">="},
  {REL_LE_TOKEN, "<="},
  {REL_NE_TOKEN, "!="},
  {RENDER_TOKEN, "render"},
  {REPEAT_TOKEN, "repeat"},
  {RGBFT_TOKEN, "rgbft"},
  {RGBF_TOKEN, "rgbf"},
  {RGBT_TOKEN, "rgbt"},
  {RGB_TOKEN, "rgb"},
  {RIGHT_ANGLE_TOKEN, ">"},
  {RIGHT_CURLY_TOKEN, "}"},
  {RIGHT_PAREN_TOKEN, ")"},
  {RIGHT_SQUARE_TOKEN, "]"},
  {RIGHT_TOKEN, "right"},
  {RIPPLES_TOKEN, "ripples"},
  {ROTATE_TOKEN, "rotate"},
  {ROUGHNESS_TOKEN, "roughness"},
  {SAMPLES_TOKEN, "samples"},
  {SCALE_TOKEN, "scale"},
  {SCALLOP_WAVE_TOKEN, "scallop_wave"},
  {SCATTERING_TOKEN, "scattering"},
  {SEED_TOKEN, "seed"},
#ifdef SignPatch
  {SGN_TOKEN, "sgn"}, /*Chris Huff jan 2001*/
#endif
  {SEMI_COLON_TOKEN, ";"},
  {SINE_WAVE_TOKEN, "sine_wave"},
  {SINGLE_QUOTE_TOKEN, "'"},
  {SINH_TOKEN, "sinh"},
  {SIN_TOKEN, "sin"},
  {SKYSPHERE_ID_TOKEN, "sky_sphere identifier"},
  {SKYSPHERE_TOKEN, "sky_sphere"},
  {SKY_TOKEN, "sky"},
  {SLASH_TOKEN, "/"},
  {SLICE_TOKEN, "slice"},
  {SLOPE_MAP_ID_TOKEN, "slope_map identifier"},
  {SLOPE_MAP_TOKEN, "slope_map"},
#ifdef ColorTrianglePatch
  {SMOOTH_COLOR_TRIANGLE_TOKEN, "smooth_color_triangle"}, /* [AP] */
#endif    
	{SMOOTH_TOKEN, "smooth"},
  {SMOOTH_TRIANGLE_TOKEN, "smooth_triangle"},
  {SOR_TOKEN, "sor"},
  {SPECULAR_TOKEN, "specular"},
  {SPHERE_TOKEN, "sphere"},
  {SPHERICAL_CAMERA_TOKEN, "spherical_camera"},
  {SPHERICAL_TOKEN, "spherical"},
  {SPIRAL1_TOKEN, "spiral1"},
  {SPIRAL2_TOKEN, "spiral2"},
#ifdef BsplinePatch
    {SPLINE_CLOSED_TOKEN, "closed"},
    {SPLINE_DIRECTION_TOKEN, "start_end_direction"},
    {SPLINE_NATURAL_TOKEN, "natural"},
#endif
#ifdef SplinePatch
  {SPLINE_ID_TOKEN, "spline identifier"},
  {SPLINE_TOKEN, "spline"},
#endif
  {SPOTLIGHT_TOKEN, "spotlight"},
  {SPOTTED_TOKEN, "spotted"},
  {SQRT_TOKEN, "sqrt"},
#ifdef TrianglulairSquarePatch 
  {SQUARE_TOKEN, "square"},
#endif
  {SQR_TOKEN, "sqr"},
  {STAR_TOKEN, "*"},
  {STATISTICS_TOKEN, "statistics"},
  {STRCMP_TOKEN, "strcmp"},
  {STRENGTH_TOKEN, "strength"},
  {STRING_ID_TOKEN, "string identifier"},
  {STRING_LITERAL_TOKEN, "string literal"},
  {STRLEN_TOKEN, "strlen"},
  {STRLWR_TOKEN, "strlwr"},
  {STRUPR_TOKEN, "strupr"},
  {STR_TOKEN, "str"},
  {STURM_TOKEN, "sturm"},
  {SUBSTR_TOKEN, "substr"},
  {SUPERELLIPSOID_TOKEN, "superellipsoid"},
  {SWITCH_TOKEN, "switch"},
  {SYS_TOKEN, "sys"},
  {TANH_TOKEN, "tanh"},
  {TAN_TOKEN, "tan"},
  {TEXTURE_ID_TOKEN, "texture identifier"},
  {TEXTURE_MAP_ID_TOKEN, "texture_map identifier"},
  {TEXTURE_MAP_TOKEN, "texture_map"},
#ifdef TrianglulairSquarePatch 
  {TRIANGULAR_TOKEN, "triangular"},
#endif
  {TEXTURE_TOKEN, "texture"},
#ifdef InteriorTexturePatch
  {INTERIOR_TEXTURE_TOKEN, "interior_texture"},/*Chris Huff "interior texture patch"*/
#endif
#if defined(ProximityPatch) || defined(BlobPatternPatch)
  {MAX_DENSITY_TOKEN, "max_density"}, /*Chris Huff blob pattern & Proximity*/
#endif
#if defined(ProximityPatch) 
  {SAMPLE_WEIGHTING_TOKEN, "sample_weighting"}, /*Chris Huff curvature pattern & Proximity*/
#endif
#ifdef BlobPatternPatch
  {BLOB_PIGMENT_TOKEN, "blob_pigment"}, /*Chris Huff blob pigment*/
  {DENSITY_FUNCTION_TOKEN, "density_function"},/*Chris Huff blob pattern*/
#endif
#ifdef ProximityPatch
  {PROXIMITY_TOKEN, "proximity"},/*Chris Huff proximity pattern*/
  {SIDES_TOKEN, "sides"},/*Chris Huff proximity pattern*/
  {SAMPLE_BAILOUT_TOKEN, "sample_bailout"},/*Chris Huff proximity pattern*/
#endif
  {TEXT_TOKEN, "text"},
  {TGA_TOKEN, "tga"},
  {THICKNESS_TOKEN, "thickness"},
  {THRESHOLD_TOKEN, "threshold"},
  {TIGHTNESS_TOKEN, "tightness"},
  {TILDE_TOKEN, "~"},
  {TILE2_TOKEN, "tile2"},
  {TILES_TOKEN, "tiles"},
  {TNORMAL_ID_TOKEN, "normal identifier"},
  {TNORMAL_TOKEN, "normal"},
  {TORUS_TOKEN, "torus"},
  {TRACK_TOKEN, "track"},
  {TRANSFORM_ID_TOKEN, "transform identifier"},
  {TRANSFORM_TOKEN, "transform"},
  {TRANSLATE_TOKEN, "translate"},
  {TRANSMIT_TOKEN, "transmit"},
  {TRIANGLE_TOKEN, "triangle"},
  {TRIANGLE_WAVE_TOKEN, "triangle_wave"},
#ifdef RBezierPatch
  {TRIMMED_BY_ID_TOKEN, "trimmed_by identifier"},
  {TRIMMED_BY_TOKEN, "trimmed_by"},
#endif 
  {TRUE_TOKEN, "true"},
  {TTF_TOKEN, "ttf"},
  {TURBULENCE_TOKEN, "turbulence"},
  {TURB_DEPTH_TOKEN, "turb_depth"},
  {TYPE_TOKEN, "type"},
  {T_TOKEN, "t"},
  {ULTRA_WIDE_ANGLE_TOKEN, "ultra_wide_angle"},
  {UNDEF_TOKEN, "undef"},
  {UNION_TOKEN, "union"},
  {UP_TOKEN, "up"},
  {USE_COLOUR_TOKEN, "use_color"},
  {USE_COLOUR_TOKEN, "use_colour"},
  {USE_INDEX_TOKEN, "use_index"},
  {UV_ID_TOKEN, "uv vector identifier"},
  {U_STEPS_TOKEN, "u_steps"},
  {U_TOKEN, "u"},
  {VAL_TOKEN, "val"},
  {V_ANGLE_TOKEN, "v_angle"}, /* MH */
  {VARIANCE_TOKEN, "variance"},
  {VAXIS_ROTATE_TOKEN, "vaxis_rotate"},
  {VCROSS_TOKEN, "vcross"},
  {VDOT_TOKEN, "vdot"},
  {VECTOR_4D_ID_TOKEN, "4d-vector identifier"},
  {VECTOR_FUNCT_TOKEN, "vector function"},
  {VECTOR_ID_TOKEN, "vector identifier"},
  {VERSION_TOKEN, "version"},
  {VLENGTH_TOKEN, "vlength"},
  {VNORMALIZE_TOKEN, "vnormalize"},
  {VROTATE_TOKEN, "vrotate"},
#ifdef ComplexFunctionsPatch
/*dfs -- complex functions*/
  {CREAL_TOKEN, "creal"},
  {CIMAG_TOKEN, "cimag"},
  {CABS2_TOKEN, "cabs2"},
  {CCONJ_TOKEN, "cconj"},
  {CSQR_TOKEN, "csqr"},
  {CSQRT_TOKEN, "csqrt"},
  {CMULT_TOKEN, "cmult"},
  {CDIV_TOKEN, "cdiv"},
  {CEXP_TOKEN, "cexp"},
  {CLOG_TOKEN, "clog"},
  {CPOW_TOKEN, "cpow"},
  {CSIN_TOKEN, "csin"},
  {CCOS_TOKEN, "ccos"},
  {CTAN_TOKEN, "ctan"},
  {CSINH_TOKEN, "csinh"},
  {CCOSH_TOKEN, "ccosh"},
  {CTANH_TOKEN, "ctanh"},
  {CASIN_TOKEN, "casin"},
  {CACOS_TOKEN, "cacos"},
  {CATAN_TOKEN, "catan"},
  {CASINH_TOKEN, "casinh"},
  {CACOSH_TOKEN, "cacosh"},
  {CATANH_TOKEN, "catanh"},
/*--dfs*/
#endif
#ifdef VturbulencePatch
  {VTURBULENCE_TOKEN, "vturbulence"}, /* Added by bajcik */
#endif  
#ifdef CracklePatch 
  {COORDS_TOKEN, "coords" },
  {FACETS_TOKEN, "facets"},
  {FORM_TOKEN, "form"},
  {METRIC_TOKEN, "metric" },
  {SIZE_TOKEN, "size" },
#endif
#if defined(CracklePatch) || defined(SolidPatternPatch)
  {SOLID_TOKEN, "solid" },
#endif
#ifdef ClothPatternPatch
  {CLOTH_TOKEN, "cloth"}, /*Chris Huff "cloth pattern"*/
  {CLOTH2_TOKEN, "cloth2"},/*Chris Huff "cloth2 pattern"*/
#endif
#ifdef TorodialPatch
  {TOROIDAL_SPIRAL_TOKEN, "torspiral"}, /*Chris Huff "toroidal spiral pattern"*/
#endif
  {V_STEPS_TOKEN, "v_steps"},
  {V_TOKEN, "v"},
  {WARNING_TOKEN, "warning"},
  {WARP_TOKEN, "warp"},
  {WATER_LEVEL_TOKEN, "water_level"},
  {WAVES_TOKEN, "waves"},
  {WHILE_TOKEN, "while"},
  {WIDTH_TOKEN, "width"},
  {WOOD_TOKEN, "wood"},
  {WRINKLES_TOKEN, "wrinkles"},
  {WRITE_TOKEN, "write"},
  {X_TOKEN, "x"},
  {YES_TOKEN, "yes"},
  {Y_TOKEN, "y"},
  {Z_TOKEN, "z"},                                
  /* NK 1998 */
  /* these are for pattern coordinate types */
  {UV_MAPPING_TOKEN, "uv_mapping"},
  /* next one for doing custom uv coordinates in various objects */
  {UV_VECTORS_TOKEN, "uv_vectors"},
  /* for mesh syntax version 2 */
  {MESH2_TOKEN, "mesh2"},
  {VERTEX_VECTORS_TOKEN, "vertex_vectors"},
  {NORMAL_VECTORS_TOKEN, "normal_vectors"},
  {FACE_INDICES_TOKEN, "face_indices"},
  {UV_INDICES_TOKEN, "uv_indices"},
  {NORMAL_INDICES_TOKEN, "normal_indices"},
  {TEXTURE_LIST_TOKEN, "texture_list"},
  {IMAGE_PATTERN_TOKEN, "image_pattern"},
  {USE_ALPHA_TOKEN, "use_alpha"},
  {RESET_CHILDREN_TOKEN, "reset_children"},
  {DOUBLE_ILLUMINATE_TOKEN, "double_illuminate"},
  {INSIDE_VECTOR_TOKEN, "inside_vector"},
  /* NK phmap */
  {PHOTONS_TOKEN, "photons"},
  {STEPS_TOKEN,"steps"},
  {PASS_THROUGH_TOKEN, "pass_through"},
  {IGNORE_PHOTONS_TOKEN, "ignore_photons"},
  {AUTOSTOP_TOKEN, "autostop"},
  {GATHER_TOKEN, "gather"},
  {SPLIT_UNION_TOKEN, "split_union"},
  {RANGE_DIVIDER_TOKEN, "range_divider"},
  {EXPAND_THRESHOLDS_TOKEN, "expand_thresholds"},
  {SPACING_TOKEN, "spacing"},
  {LOAD_FILE_TOKEN, "load_file"},
  {SAVE_FILE_TOKEN, "save_file"},
  {GLOBAL_TOKEN, "global"},
  {TARGET_TOKEN, "target"},
  {INI_OPTION_TOKEN, "ini_option"},
  /* NK ---- */
  /* NK rad */
  {MAX_SAMPLE_TOKEN, "max_sample"},
  {PRETRACE_START_TOKEN, "pretrace_start"},
  {PRETRACE_END_TOKEN, "pretrace_end"},
  {ALWAYS_SAMPLE_TOKEN, "always_sample"},
  /* NK ---- */
  {AA_LEVEL_TOKEN, "aa_level"},
  {AA_THRESHOLD_TOKEN, "aa_threshold"},
  {METHOD_TOKEN, "method"}, /*MH */
  {FRACTAL_EXTERIOR_TYPE_TOKEN, "fractal_exterior_type"},
  {FRACTAL_INTERIOR_TYPE_TOKEN, "fractal_interior_type"},
  {MAGNET1J_TOKEN, "magnet1j"},
  {MAGNET1M_TOKEN, "magnet1m"},
  {MAGNET2J_TOKEN, "magnet2j"},
  {MAGNET2M_TOKEN, "magnet2m"},
  {MANDEL3_TOKEN, "mandel3"},
  {MANDEL4_TOKEN, "mandel4"},
  {JULIA_TOKEN, "julia"},
  {JULIA3_TOKEN, "julia3"},
  {JULIA4_TOKEN, "julia4"},
  /* NK ---- */
  /* NK persist */
  {CREATE_TOKEN, "create"},
  {LABEL_TOKEN, "label"},
  {MODIFY_TOKEN, "modify"},
  {DESTROY_TOKEN, "destroy"},
  /* NK ---- */
  /* NK misc */
  {DISP_NELEMS_TOKEN, "disp_nelems"},
  {DISPERSION_TOKEN, "dispersion"},
  /* NK ---- */
  /* NK super */
  {B_SPLINE_SPHERE_SWEEP_TOKEN, "b_spline_sphere_sweep"},       
  {CATMULL_ROM_SPLINE_SPHERE_SWEEP_TOKEN, "catmull_rom_spline_sphere_sweep"},
  {DIM_TOKEN, "dim" },
  {EQUPOTSURFACE_TOKEN,"equpotsurface"},
  {FLOAT_W_TOKEN, "float"},
  {LINEAR_SPHERE_SWEEP_TOKEN, "linear_sphere_sweep"},
#ifdef PatternPatch
  {PATTERN_TOKEN, "pattern"},
#endif  
#if defined(Noise3DPatch) || defined(POVISO)
  {NOISE3D_TOKEN,"noise3d"},
#endif
   {ACCURACY_TOKEN, "accuracy"}, /*is now part of MegaPov (normal)*/
#ifdef POVISO
  {MAX_GRADIENT_TOKEN, "max_gradient"},
  {MAX_TRACE_TOKEN, "max_trace"},
  {FUNC_TOKEN,"func"},
  {FUNCTION_TOKEN,"function"},
  {NOEVAL_TOKEN,"no_eval"},
  {EVAL_TOKEN,"eval"},
  {LIBNAME_TOKEN,"lib_name"},
  {LIBRARY_TOKEN,"library"},
  {FUNC_ID_TOKEN,"func_id"},
  {R_TOKEN,"r"},
  {S_TOKEN,"s"},
  {LN_TOKEN,"ln"},
  {CUB_TOKEN,"cub"},
  {FUNC3D_TOKEN, "func3d"},
  {ALL_INTERSECTIONS_TOKEN,"all_intersections"},
  {PARAMETRIC_TOKEN,"parametric"},
  {ISOSURFACE_TOKEN,"isosurface"},
  {CLOSE_TOKEN,"close"},
  {SIGN_TOKEN,"sign"},
  {MAX_GRAD_TOKEN, "max_grad"},
  {PRECOMPUTE_TOKEN,"precompute"}, /* Fri 09-27-1996 O. */
  {CONTAINED_BY_TOKEN,"contained_by"},
#endif
  {REFLECTION_BLUR_TOKEN, "reflection_blur"},
  {REFLECTION_BLUR_MAX_TOKEN, "reflection_blur_max"},
  {REFLECTION_BLUR_MAX_ADC_TOKEN, "reflection_blur_max_adc"},
  {REFLECTION_FALLOFF_TOKEN, "reflection_falloff"},
  {REFLECTION_MAX_TOKEN, "reflection_max"},
  {REFLECT_METALLIC_TOKEN, "reflect_metallic"},
  {REFLECTION_MIN_TOKEN, "reflection_min"},
  {REFLECTION_SAMPLES_TOKEN, "reflection_samples"},
  {REFLECTION_TYPE_TOKEN, "reflection_type"},
  {SLOPE_TOKEN,"slope"},
  {SPHERE_SWEEP_DEPTH_TOLERANCE_TOKEN, "sphere_sweep_depth_tolerance"},
  {SPHERE_SWEEP_TOKEN, "sphere_sweep"},
  {STRING_TOKEN, "string"},
  {SYSTEM_TOKEN, "system"},       /* added djconnel 12Sep1998 */
#ifdef IsoBlobPatch
/* begin Lummox JR's added tokens */
  {ISOBLOB_TOKEN, "isoblob"},
/* end Lummox JR's added tokens */
#endif
#ifdef CellsPatch
  {CELLS_TOKEN,"cells"},
#endif
#ifdef VanSicklePatternPatch
  {BLOTCHES_TOKEN,"blotches"},
  {BANDS_TOKEN,"bands"},
  {SHEET_TOKEN,"sheet"},
#endif
#ifdef PatternWarpPatch
  {TOROIDAL_TOKEN, "toroidal"},
  {ORIENTATION_TOKEN, "orientation"},
  {DIST_EXP_TOKEN, "dist_exp"},
  {MAJOR_RADIUS_TOKEN, "major_radius"},
#endif
#ifdef TracePatch
  {TRACE_TOKEN, "trace"},
#endif  
#ifdef MotionBlurPatch
  {MOTION_BLUR_TOKEN, "motion_blur"},
  {MOTION_BLUR2_TOKEN, "motion_blur2"},
#endif
#ifdef BlinnPatch
  {BLINN_TOKEN, "blinn"},
#endif
#ifdef NormalBugFix
  {NO_BUMP_SCALE_TOKEN, "no_bump_scale"},
#endif
#ifdef PostProcessPatch
#ifdef PostProcessKeepDataPatch
  {KEEP_DATA_FILE_TOKEN, "keep_data_file"},
#endif
  {POST_PROCESS_TOKEN, "post_process"},
  {OVERWRITE_FILE_TOKEN, "overwrite_file"},
  {FOCAL_BLUR_TOKEN, "focal_blur"},
  {DEPTH_TOKEN, "depth"},
  {SOFT_GLOW_TOKEN, "soft_glow"},

#ifdef PostProcessColorMatrixPatch
  {COLOR_MATRIX_TOKEN, "color_matrix"},
#endif
#ifdef PostProcessBlurMatrixPatch
  {CONVOLUTION_MATRIX_TOKEN, "convolution_matrix"},
#endif
#ifdef PostProcessStepsPatch
  {POSTERIZE_TOKEN,"posterize"},
#endif
#ifdef PostProcessFindEdgesPatch
  {FIND_EDGES_TOKEN, "find_edges"},
#endif
#ifdef PostProcessInvertPatch
  {INVERT_TOKEN, "invert"},
#endif
#ifdef PostProcessClipColorsPatch
  {CLIP_COLORS_TOKEN, "clip_colors"},
#endif
#ifdef PostProcessMinPatch
  {POST_MIN_TOKEN, "post_min"},
#endif
#ifdef PostProcessMaxPatch
  {POST_MAX_TOKEN, "post_max"},
#endif
#ifdef PostProcessAddPatch
  {ADD_TOKEN, "add"},
#endif
#ifdef PostProcessSubtractPatch
  {SUBTRACT_TOKEN, "subtract"},
#endif
#ifdef PostProcessDividePatch
  {DIVIDE_TOKEN, "divide"},
#endif
#ifdef PostProcessMultiplyPatch
  {MULTIPLY_TOKEN, "multiply"},
#endif
#ifdef PostProcessExponentPatch
  {EXPONENT_TOKEN, "exponent"},
#endif
#ifdef PostProcessStarsPatch
  {STARS_TOKEN, "stars"},
#endif
#ifdef PostProcessPatternBlurPatch
  {PATTERN_BLUR_TOKEN, "patterned_blur"},
#endif
#ifdef NoisePigmentPatch
  {NOISE_PIGMENT_TOKEN, "noise_pigment"},
#endif
#ifdef DisplaceWarpPatch
  {DISPLACE_TOKEN, "displace"},
#endif
#endif


#ifdef SplineWavePatch
  {SPLINE_WAVE_TOKEN, "spline_wave"},
#endif
#ifdef PigmentPatternPatch
  {PIGMENT_PATTERN_TOKEN,"pigment_pattern"},
#endif
#ifdef MultiTextureCsgPatch
  {CUTAWAY_TEXTURES_TOKEN, "cutaway_textures"},
#endif
#ifdef SetPatch
  {SET_TOKEN, "set"},
#endif
#ifdef GlowPatch
  {GLOW_TOKEN, "glow"},
#endif  
  {CONSERVE_ENERGY_TOKEN, "conserve_energy"},
  {VECTOR_TOKEN, "vector"},
  {LIGHT_GROUP_TOKEN, "light_group"},
  {GROUPS_TOKEN, "groups"},
  {PARALLEL_TOKEN, "parallel"},
  {PROJECTED_THROUGH_TOKEN, "projected_through"},
  {STRING_ENCODING_TOKEN, "string_encoding"}
  /* NK ---- */
				/* JB +++ */
#ifdef SampleSpacingPatch
  ,
  {SAMPLE_SPACING_TOKEN, "sample_spacing"},
  {SAMPLE_JITTER_TOKEN, "sample_jitter"}
#endif
#ifdef EmissionMethodPatch
  ,
  {EMISSION_TYPE_TOKEN, "emission_type"},
  {EMISSION_EXTINCTION_TOKEN, "emission_extinction"}
#endif
				/* JB--- */

};



#ifdef FastMacroPatch
static void Echo_ungetc (int c);
#else
static int Echo_ungetc (int c);
#endif
static int Echo_getc (void);
static int Skip_Spaces (void);
static int Parse_C_Comments (void);
static void Begin_String (void);
static void Stuff_Character (int c);
static void End_String (void);
static int Read_Float (void);
static void Parse_String_Literal (void);
static void Read_Symbol (void);
#ifndef SetPatch
	static SYM_ENTRY *Find_Symbol (int Index, char *s);
#endif
static void Skip_Tokens (COND_TYPE cond);

static int get_hash_value (char *s);
static void Write_Token (TOKEN Token_Id);
static void Destroy_Table (int index);
static void init_sym_tables (void);
static void Add_Sym_Table (char *s);
static void Remove_Symbol (int Index, char *Name);
static POV_MACRO *Parse_Macro(void);
static void Invoke_Macro(void);
static void Return_From_Macro(void);
static void Add_Entry (int Index,SYM_ENTRY *Table_Entry);
static void Parse_Initalizer (int Sub, int Base, POV_ARRAY *a);      

static void Parse_Fopen(void);
static void Parse_Fclose(void);
static void Parse_Read(void);
static void Parse_Write(void);
static int Parse_Comma_RParen(void);
static int Parse_Read_Value(DATA_FILE *User_File,int Previous,int *NumberPtr,void **DataPtr);
static void Check_Macro_Vers(void);
static DBL Parse_Cond_Param(void);
static void Parse_Cond_Param2(DBL *V1,DBL *V2);
static void Inc_CS_Index(void);

/* NK 1999 - use $ for #declare */
static int Parse_Assignment();


#ifdef FastMacroPatch
static void Fopen_Close(DATA_FILE *File); /* Close the data file */
static int Fopen_Getc(DATA_FILE *File); /* Read a character */
static void Fopen_Ungetc(DATA_FILE *File, int Ch); /* Stuff a character back into the file */
static void Fopen_Putc(DATA_FILE *File, int Ch);
static int Fopen_Seek (DATA_FILE *File, int Offset);
static int Fopen_Tell (DATA_FILE *File);
static void Fopen_Setup (DATA_FILE *File);

static void String_Close(DATA_FILE *File); /* Close the data file */
static int String_Getc(DATA_FILE *File); /* Read a character */
static void String_Ungetc(DATA_FILE *File, int Ch); /* Stuff a character back into the file */
static void String_Putc(DATA_FILE *File, int Ch);
static int String_Seek (DATA_FILE *File, int Offset);
static int String_Tell (DATA_FILE *File);

static void Data_File_Printf (DATA_FILE *File, char *str, ...);
#endif

/*****************************************************************************
*
* FUNCTION
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
* DESCRIPTION
*
* CHANGES
*
******************************************************************************/

/* NK persist - Nov 1999  */
int alreadyInitializedSymTables;

void Initialize_Tokenizer()
{
  char b[FILE_NAME_LENGTH];
  Stage = STAGE_TOKEN_INIT;

  pre_init_tokenizer ();

  Temp_Data_File_For_Fread = NULL;
#ifdef FastMacroPatch
  Fopen_Setup (Data_File);
#endif

  if (opts.Options & FROM_STDIN)
  {
#ifdef FastMacroPatch
    Data_File->Data = stdin;
#else
    Data_File->File = stdin;
#endif
  }
  else
  {
     if (input_file_in_memory)
     {
        /* Note platforms which use this feature will
         * trap fopen so the NULL is ok [C.Cason 7/3/96]
         */
#ifdef FastMacroPatch
        Data_File->Data = fopen (NULL, "rt") ;
#else
        Data_File->File = fopen (NULL, "rt") ;
#endif
     }
     else
     {
#ifdef FastMacroPatch
        Data_File->Data = Locate_File (opts.Input_File_Name, READ_BINFILE_STRING,".pov",".POV",b,TRUE);
#else
        Data_File->File = Locate_File (opts.Input_File_Name, READ_TXTFILE_STRING,".pov",".POV",b,TRUE);
#endif
        strcpy(opts.Input_File_Name,b);
     }
  }

#ifdef FastMacroPatch
  if (Data_File->Data == NULL)
#else
  if (Data_File->File == NULL)
#endif
  {
    Error ("Cannot open input file.");
  }

  /* NK should this 'if' be here? */
  if (!input_file_in_memory)  /*R.S. Nov.8,96*/
  {
    Data_File->Filename = (char *)POV_MALLOC(strlen(opts.Input_File_Name)+1, "filename");

    strcpy (Data_File->Filename, opts.Input_File_Name);
  }

  Data_File->Line_Number = 0;
  Data_File->R_Flag = FALSE;
#ifdef UnofficialBlocking
  Data_File->unofficialVersion = -1;
  opts.unofficialVersion = Data_File->unofficialVersion;
#endif

  /* Init echo buffer. */

  Echo_Buff = (char **)POV_MALLOC(sizeof(char *) * Num_Echo_Lines, "echo buffer");

  for (Echo_Line = 0; Echo_Line < Num_Echo_Lines; Echo_Line++)
  {
    Echo_Buff[Echo_Line] = (char *)POV_MALLOC((size_t)Echo_Line_Length+10, "echo buffer");

    Echo_Buff[Echo_Line][0]='\0';
  }

  Echo_Line = 0;
  Echo_Ptr = Echo_Buff[0];
  Got_EOF  = FALSE;

  /* Init conditional stack. */

  Cond_Stack = (CS_ENTRY*)POV_MALLOC(sizeof(CS_ENTRY) * COND_STACK_SIZE, "conditional stack");

  Cond_Stack[0].Cond_Type    = ROOT_COND;
  Cond_Stack[0].Switch_Value = 0.0;

  /* NK persist - Nov 1999 - opts.persistentGlobals */
  if(!opts.persistentAnimation
      || !alreadyInitializedSymTables)
  {
    init_sym_tables();
    alreadyInitializedSymTables = TRUE;
  }

  Max_Trace_Level = 5;      

}



/*****************************************************************************
*
* FUNCTION
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
* DESCRIPTION
*
* CHANGES
*
******************************************************************************/

void pre_init_tokenizer ()
{
  Token.Token_Line_No = 0;
  Token.Token_String  = NULL;
  Token.Unget_Token   = FALSE;
  Token.End_Of_File   = FALSE;
  Token.Filename      = NULL;
  Token.Data = NULL;

  line_count = 10;
  token_count = 0;
  Include_File_Index = 0;
  Echo_Indx=0;
  Echo_Line=0;
  Echo_Ptr=NULL;
  Echo_Buff=NULL;

  CS_Index            = 0;
  Skipping            = FALSE;
  Inside_Ifdef        = FALSE;
  Inside_MacroDef     = FALSE;
  Cond_Stack          = NULL;
  Data_File = &Include_Files[0];
  Data_File->Filename = NULL;

  /* NK persist - Nov 1999 - opts.persistentGlobals */
  if(!opts.persistentAnimation
      || !alreadyInitializedSymTables)
  {
    Table_Index         = -1;
  }
}


/*****************************************************************************
*
* FUNCTION
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
* DESCRIPTION
*
* CHANGES
*
******************************************************************************/

void Terminate_Tokenizer(int freeEverything)
{
  /* NK persist - Nov 1999 - only destroy if not persistent
  */
  if(!opts.persistentAnimation || 
     freeEverything)
  {
    while(Table_Index >= 0)
    {
       Destroy_Table(Table_Index--);
    }
    alreadyInitializedSymTables = FALSE;
  }

#ifdef UnofficialBlocking
  /* unofficial version used for rendering */
  opts.unofficialVersion = Data_File->unofficialVersion;
#endif


  if (Temp_Data_File_For_Fread)
  {
    Data_File = Temp_Data_File_For_Fread;
    Temp_Data_File_For_Fread = NULL;
  }
  
  if (Data_File->Filename != NULL)
  {
#ifdef FastMacroPatch
    Data_File_Close (Data_File);
#else
    if (Data_File->File != NULL)
      fclose (Data_File->File);
#endif
    Got_EOF=FALSE;

    POV_FREE (Data_File->Filename);

    Data_File->Filename = NULL;
  }

  while (Include_File_Index >= 0)
  {
    Data_File = &Include_Files[Include_File_Index--];
  
    if (Data_File->Filename != NULL)
    {
#ifdef FastMacroPatch
      Data_File_Close (Data_File);
#else
      if (Data_File->File != NULL)
        fclose (Data_File->File);
#endif
      Got_EOF=FALSE;

      POV_FREE (Data_File->Filename);

      Data_File->Filename = NULL;
    }
  }

  if (Echo_Buff != NULL)
  {
    for (Echo_Line = 0; Echo_Line < Num_Echo_Lines; Echo_Line++)
    {
      if (Echo_Buff[Echo_Line]!=NULL)
      {
        POV_FREE (Echo_Buff[Echo_Line]);

        Echo_Buff[Echo_Line]=NULL;
      }
    }

    POV_FREE (Echo_Buff);

    Echo_Buff = NULL;
  }

  if (Cond_Stack!=NULL)
  {
    POV_FREE (Cond_Stack);

    Cond_Stack = NULL;
  }
}



/*****************************************************************************
*
* FUNCTION
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
* DESCRIPTION
*
*   The main tokenizing routine.  Set up the files and continue parsing
*   until the end of file
*
*   Read a token from the input file and store it in the Token variable.
*   If the token is an INCLUDE token, then set the include file name and
*   read another token.
*
*   This function performs most of the work involved in tokenizing.  It
*   reads the first character of the token and decides which function to
*   call to tokenize the rest.  For simple tokens, it simply writes them
*   out to the token buffer.
*
* CHANGES
*
******************************************************************************/

void Get_Token ()
{
  register int c,c2;

  if (Token.Unget_Token)
  {
    Token.Unget_Token = FALSE;

    return;
  }

  if (Token.End_Of_File)
  {
    return;
  }

  Token.Token_Id = END_OF_FILE_TOKEN;

  while (Token.Token_Id == END_OF_FILE_TOKEN)
  {
    Skip_Spaces();

    c = Echo_getc();

    if (c == EOF)
    {
      if (Data_File->R_Flag)
      {
        Token.Token_Id = END_OF_FILE_TOKEN;
        Token.End_Of_File = TRUE;
        return;
      }
      
      if (Include_File_Index == 0)
      {
        if (CS_Index !=0)
          Error("End of file reached but #end expected.");

        Token.Token_Id = END_OF_FILE_TOKEN;

        Token.End_Of_File = TRUE;

        Status_Info("\n");
        
        return;
      }

      UICB_CLOSE_INCLUDE_FILE  /* Notify UI that we are about to close an include file normally */

#ifdef FastMacroPatch
      Data_File_Close (Data_File);
#else
      fclose(Data_File->File); /* added to fix open file buildup JLN 12/91 */
#endif
      Got_EOF=FALSE;
      
      Destroy_Table(Table_Index--);

      POV_FREE (Data_File->Filename);

      Data_File = &Include_Files[--Include_File_Index];

#ifdef UnofficialBlocking
      opts.unofficialVersion = Data_File->unofficialVersion;
#endif

      continue;
    }

    String[0] = c; /* This isn't necessar but helps debugging */

    String[1] = '\0';

    String_Index = 0;

    switch (c)
    {
      case '\n':
        Data_File->Line_Number++;
        COOPERATE_0
        break;

      case '{' :
        Write_Token (LEFT_CURLY_TOKEN);
        break;
  
      case '}' :
        Write_Token (RIGHT_CURLY_TOKEN);
        break;
  
      case '@' :
        Write_Token (AT_TOKEN);
        break;
  
      case '&' :
        Write_Token (AMPERSAND_TOKEN);
        break;
  
      case '`' :
        Write_Token (BACK_QUOTE_TOKEN);
        break;
  
      case '\\':
        Write_Token (BACK_SLASH_TOKEN);
        break;

      case '|' :
        Write_Token (BAR_TOKEN);
        break;
  
      case ':' :
        Write_Token (COLON_TOKEN);
        break;
  
      case ',' :
        Write_Token (COMMA_TOKEN);
        break;
  
      case '-' :
        Write_Token (DASH_TOKEN);
        break;
  
      case '$' :
        Write_Token (DOLLAR_TOKEN);
        /* NK 1999 - use $ for #declare */
        if(!Ok_To_Declare)
        {
          Token.Token_Id=DOLLAR_TOKEN; 
          Token.Unget_Token = FALSE;
        }
        else
          Parse_Assignment();
        /* NK ---- */
        break;
  
      case '%' :
        Write_Token (PERCENT_TOKEN);
        /* NK 1999 - use $ for #declare */
        if(!Ok_To_Declare)
        {
          Token.Token_Id=PERCENT_TOKEN; 
          Token.Unget_Token = FALSE;
        }
        else
          Parse_Assignment();
        /* NK ---- */
        break;
  
      case '=' :
        Write_Token (EQUALS_TOKEN);
        break;
  
      case '!' :
        c2 = Echo_getc();
        if (c2 == (int)'=')
        {
          Write_Token (REL_NE_TOKEN);
        }
        else
        {
          Echo_ungetc(c2);
          Write_Token (EXCLAMATION_TOKEN);
        }
        break;
  
      case '#' : 
        Parse_Directive(TRUE);
        /* Write_Token (HASH_TOKEN);*/
        break;
  
      case '^' :
        Write_Token (HAT_TOKEN);
        break;
  
      case '<' :
        c2 = Echo_getc();
        if (c2 == (int)'=')
        {
          Write_Token (REL_LE_TOKEN);
        }
        else
        {
          Echo_ungetc(c2);
          Write_Token (LEFT_ANGLE_TOKEN);
        }
        break;
  
      case '(' :
        Write_Token (LEFT_PAREN_TOKEN);
        break;
  
      case '[' :
        Write_Token (LEFT_SQUARE_TOKEN);
        break;
  
      case '+' :
        Write_Token (PLUS_TOKEN);
        break;
  
      case '?' :
        Write_Token (QUESTION_TOKEN);
        break;
  
      case '>' :
        c2 = Echo_getc();
        if (c2 == (int)'=')
        {
          Write_Token (REL_GE_TOKEN);
        }
        else
        {
          Echo_ungetc(c2);
          Write_Token (RIGHT_ANGLE_TOKEN);
        }
        break;
  
      case ')' :
        Write_Token (RIGHT_PAREN_TOKEN);
        break;
  
      case ']' :
        Write_Token (RIGHT_SQUARE_TOKEN);
        break;
  
      case ';' : /* Parser doesn't use it, so let's ignore it */
        Write_Token (SEMI_COLON_TOKEN);
        break;
  
      case '\'':
        Write_Token (SINGLE_QUOTE_TOKEN);
        break;
  
        /* enable C++ style commenting */
      case '/' :
        c2 = Echo_getc();
        if(c2 != (int) '/' && c2 != (int) '*')
        {
          Echo_ungetc(c2);
          Write_Token (SLASH_TOKEN);
          break;
        }
        if(c2 == (int)'*')
        {
          Parse_C_Comments();
          break;
        }
        while((c2 != (int)'\n') && (c2 != (int)'\r'))
        {
          c2=Echo_getc();
          if(c2==EOF)
          {
            Echo_ungetc(c2);
            break;
          }
        }
        if (c2 =='\n')
        {
          Data_File->Line_Number++;
        }
        COOPERATE_0
        break;

      case '*' :
        Write_Token (STAR_TOKEN);
        break;
  
      case '~' :
        Write_Token (TILDE_TOKEN);
        break;
  
      case '"' :
        Parse_String_Literal ();
        break;
  
      case '0':
      case '1':
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':
      case '8':
      case '9':
      case '.':
        Echo_ungetc(c);
        if (Read_Float () != TRUE)
          return;
        break;
  
      case 'a':
      case 'b':
      case 'c':
      case 'd':
      case 'e':
      case 'f':
      case 'g':
      case 'h':
      case 'i':
      case 'j':
      case 'k':
      case 'l':
      case 'm':
      case 'n':
      case 'o':
      case 'p':
      case 'q':
      case 'r':
      case 's':
      case 't':
      case 'u':
      case 'v':
      case 'w':
      case 'x':
      case 'y':
      case 'z':
  
      case 'A':
      case 'B':
      case 'C':
      case 'D':
      case 'E':
      case 'F':
      case 'G':
      case 'H':
      case 'I':
      case 'J':
      case 'K':
      case 'L':
      case 'M':
      case 'N':
      case 'O':
      case 'P':
      case 'Q':
      case 'R':
      case 'S':
      case 'T':
      case 'U':
      case 'V':
      case 'W':
      case 'X':
      case 'Y':
      case 'Z':
      case '_':
        Echo_ungetc(c);
        Read_Symbol ();
        break;
      case '\t':
      case '\r':
      case '\032':   /* Control Z - EOF on many systems */
      case '\0':
        break;
  
      default:
        Error("Illegal character in input file, value is %02x.\n", c);
        break;
    }
  }

  token_count++;

  if (token_count > 1000)
  {
    token_count = 0;

    COOPERATE_0

    Check_User_Abort(FALSE);

    Status_Info(".");

    line_count++;

    if (line_count > 78)
    {
      line_count = 0;

      Status_Info ("\n");
    }
  }
}



/*****************************************************************************
*
* FUNCTION
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
* DESCRIPTION
*
*   Mark that the token has been put back into the input stream.  The next
*   call to Get_Token will return the last-read token instead of reading a
*   new one from the file.
*
* CHANGES
*
******************************************************************************/

void Unget_Token ()
{
  Token.Unget_Token = TRUE;
}



/*****************************************************************************
*
* FUNCTION
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
* DESCRIPTION
*
*   Skip over spaces in the input file.
*
* CHANGES
*
******************************************************************************/

static int Skip_Spaces ()
{
  register int c;

  while (TRUE)
  {
    c = Echo_getc();

    if (c == EOF)
    {
      return (FALSE);
    }

    if (!(isspace(c) || c == 0x0A))
    {
      break;
    }

    if (c == '\n')
    {
      Data_File->Line_Number++;

      COOPERATE_0
    }
  }

  Echo_ungetc(c);

  return (TRUE);
}



/*****************************************************************************
*
* FUNCTION
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
* DESCRIPTION
*
*   C style comments with asterik and slash - CdW 8/91.
*
* CHANGES
*
******************************************************************************/

static int Parse_C_Comments()
{
  register int c, c2;
  int End_Of_Comment;

  End_Of_Comment = FALSE;

  while (!End_Of_Comment)
  {
    c = Echo_getc();

    if (c == EOF)
    {
      Error ("No */ closing comment found.");
    }

    if (c == (int) '\n')
    {
      Data_File->Line_Number++;

      COOPERATE_0
    }

    if (c == (int) '*')
    {
      c2 = Echo_getc();

      if (c2 != (int) '/')
      {
        Echo_ungetc(c2);
      }
      else
      {
        End_Of_Comment = TRUE;
      }
    }

    /* Check for and handle nested comments */

    if (c == (int) '/')
    {
      c2 = Echo_getc();

      if (c2 != (int) '*')
      {
        Echo_ungetc(c2);
      }
      else
      {
        Parse_C_Comments();
      }
    }
  }

  return (TRUE);
}



/* The following routines make it easier to handle strings.  They stuff
   characters into a string buffer one at a time making all the proper
   range checks.  Call Begin_String to start, Stuff_Character to put
   characters in, and End_String to finish.  The String variable contains
   the final string. */

/*****************************************************************************
*
* FUNCTION
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
* DESCRIPTION
*
* CHANGES
*
******************************************************************************/

static void Begin_String()
{
  String_Index = 0;
}



/*****************************************************************************
*
* FUNCTION
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
* DESCRIPTION
*
* CHANGES
*
******************************************************************************/

static void Stuff_Character(int c)
{
  if (String_Index < MAX_STRING_INDEX)
  {
    String[String_Index++] = (char)c;

    if (String_Index >= MAX_STRING_INDEX)
    {
      Error ("String too long.");
    }
  }
}



/*****************************************************************************
*
* FUNCTION
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
* DESCRIPTION
*
* CHANGES
*
******************************************************************************/

static void End_String()
{
  Stuff_Character((int)'\0');
}



/*****************************************************************************
*
* FUNCTION
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
* DESCRIPTION
*
*   Read a float from the input file and tokenize it as one token. The phase
*   variable is 0 for the first character, 1 for all subsequent characters
*   up to the decimal point, 2 for all characters after the decimal
*   point, 3 for the E+/- and 4 for the exponent.  This helps to insure
*   that the number is formatted properly. E format added 9/91 CEY
*
* CHANGES
*
******************************************************************************/

static int Read_Float()
{
  register int c, Finished, Phase;

  Finished = FALSE;

  Phase = 0;

  Begin_String();

  while (!Finished)
  {
    c = Echo_getc();

    if (c == EOF)
    {
      Error ("Unexpected end of file.");
    }

    switch (Phase)
    {
      case 0:

        Phase = 1;

        if (isdigit(c))
        {
          Stuff_Character(c);
        }
        else
        {
          if (c == '.')
          {
            c = Echo_getc();

            if (c == EOF)
            {
              Error ("Unexpected end of file");
            }

            if (isdigit(c))
            {
              Stuff_Character('0');
              Stuff_Character('.');
              Stuff_Character(c);

              Phase = 2;
            }
            else
            {
              Echo_ungetc(c);

              Write_Token (PERIOD_TOKEN);

              return(TRUE);
            }
          }
          else
          {
            Error ("Error in decimal number");
          }
        }

        break;

      case 1:
        if (isdigit(c))
        {
          Stuff_Character(c);
        }
        else
        {
          if (c == (int) '.')
          {
            Stuff_Character(c); Phase = 2;
          }
          else
          {
            if ((c == 'e') || (c == 'E'))
            {
              Stuff_Character(c); Phase = 3;
            }
            else
            {
              Finished = TRUE;
            }
          }
        }

        break;

      case 2:

        if (isdigit(c))
        {
          Stuff_Character(c);
        }
        else
        {
          if ((c == 'e') || (c == 'E'))
          {
            Stuff_Character(c); Phase = 3;
          }
          else
          {
            Finished = TRUE;
          }
        }

        break;

      case 3:

        if (isdigit(c) || (c == '+') || (c == '-'))
        {
          Stuff_Character(c); Phase = 4;
        }
        else
        {
          Finished = TRUE;
        }

        break;

      case 4:

        if (isdigit(c))
        {
          Stuff_Character(c);
        }
        else
        {
          Finished = TRUE;
        }

        break;
    }
  }

  Echo_ungetc(c);

  End_String();

  Write_Token (FLOAT_TOKEN);

  if (sscanf (String, DBL_FORMAT_STRING, &Token.Token_Float) == 0)
  {
    return (FALSE);
  }

  return (TRUE);
}



/*****************************************************************************
*
* FUNCTION
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
* DESCRIPTION
*
*   Parse a string from the input file into a token.
*
* CHANGES
*
******************************************************************************/

static void Parse_String_Literal()
{
  register int c;

  Begin_String();

  while (TRUE)
  {
    c = Echo_getc();

    if (c == EOF)
    {
      Error ("No end quote for string.");
    }

    if (c == '\\')
    {
      switch(c = Echo_getc())
      {
        case '\n':
        case '\r':

          Error("Unterminated string literal.");

          break;

        case '\"':

          c = 0x22;

          break;

        case EOF:

          Error ("No end quote for string.");

          break;

        case '\\' :
          c='\\';
          break;

        default:

          Stuff_Character ('\\');
      }

      Stuff_Character (c);
    }
    else
    {
      if (c != (int) '"')
      {
        Stuff_Character (c);
      }
      else
      {
        break;
      }
    }
  }

  End_String();

  Write_Token (STRING_LITERAL_TOKEN);

  Token.Token_String = String;
}



/*****************************************************************************
*
* FUNCTION
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
* DESCRIPTION
*
*   Read in a symbol from the input file. Check to see if it is a reserved
*   word. If it is, write out the appropriate token. Otherwise, write the
*   symbol out to the symbol table and write out an IDENTIFIER token. An
*   identifier token is a token whose token number is greater than the
*   highest reserved word.
*
* CHANGES
*
******************************************************************************/

static void Read_Symbol()
{
  register int c;
  int Local_Index,i,j,k;
  POV_ARRAY *a;
  SYM_ENTRY *Temp_Entry;
  POV_PARAM *Par;
  DBL val;

  Begin_String();

  while (TRUE)
  {
    c = Echo_getc();

    if (c == EOF)
    {
      Error ("Unexpected end of file.");
    }

    if (isalpha(c) || isdigit(c) || c == (int) '_')
    {
      Stuff_Character(c);
    }
    else
    {
      Echo_ungetc(c);

      break;
    }
  }

  End_String();

  if (Inside_Ifdef)
  {
    Token.Token_Id = IDENTIFIER_TOKEN;

    return;
  }

  /* If its a reserved keyword, write it and return */
  if ( (Temp_Entry = Find_Symbol(0,String)) != NULL)
  {
    Write_Token (Temp_Entry->Token_Number);
    return;
  }
  
  if (!Skipping)
  {
    /* Search tables from newest to oldest */
    for (Local_Index=Table_Index; Local_Index > 0; Local_Index--)
    {
      if ((Temp_Entry = Find_Symbol(Local_Index,String)) != NULL)
      {
         /* Here its a previously declared identifier. */
    
         if (Temp_Entry->Token_Number==MACRO_ID_TOKEN)
         {
           Token.Data = Temp_Entry->Data;
           if (Ok_To_Declare)
           {
              Invoke_Macro();
           }
           else
           {
              Token.Token_Id=MACRO_ID_TOKEN;
           }
           return;
         }
         
         Token.Token_Id  =   Temp_Entry->Token_Number;
         Token.NumberPtr = &(Temp_Entry->Token_Number);
         Token.DataPtr   = &(Temp_Entry->Data);
  
         while ((Token.Token_Id==PARAMETER_ID_TOKEN) ||
                (Token.Token_Id==ARRAY_ID_TOKEN))
         {
           if (Token.Token_Id==ARRAY_ID_TOKEN)
           {
             Skip_Spaces();
             c = Echo_getc();
             Echo_ungetc(c);
             
             if (c!='[')
             {
               break;
             }
             
             a = (POV_ARRAY *)(*(Token.DataPtr));
             j = 0;
  
             for (i=0; i <= a->Dims; i++)
             {
                GET(LEFT_SQUARE_TOKEN)
                val=Parse_Float();
                if (val<0.0)
                {
                  Error("Negative subscript");
                }

                k=(int)(1.0e-08+val);

                if (k>=a->Sizes[i])
                {
                   Error("Array subscript out of range");
                }
                j += k * a->Mags[i];
                GET(RIGHT_SQUARE_TOKEN)
             }
             
             Token.DataPtr   = &(a->DataPtrs[j]);
             Token.NumberPtr = &(a->Type);
             Token.Token_Id = a->Type;
             if (!LValue_Ok)
             {
                if (*Token.DataPtr==NULL)
                {
                  Error("Attempt to access uninitialized array element.");
                }
             }
           }
           else
           {
             Par             = (POV_PARAM *)(Temp_Entry->Data);
             Token.Token_Id  = *(Par->NumberPtr);
             Token.NumberPtr = Par->NumberPtr;
             Token.DataPtr   = Par->DataPtr;
           }
         }
  
         Write_Token (Token.Token_Id);
       
         Token.Data        = *(Token.DataPtr);
         Token.Table_Index = Local_Index;
         return;
      }
    }
  }

  Write_Token(IDENTIFIER_TOKEN);
}

void Write_Token (TOKEN Token_Id)
{
   Token.Token_Line_No = Data_File->Line_Number;
   Token.Filename      = Data_File->Filename;
   Token.Token_String  = String;
   Token.Data          = NULL;
   Token.Token_Id      = Token_Id;
/** poviso: July 15, '96 R.S. **/
#ifdef POVISO
   if (Func_Parse_Flag ==TRUE) return;
#endif
/** --- **/

   Token.Function_Id = Token.Token_Id;
   if (Token.Token_Id < FLOAT_FUNCT_TOKEN)
   {
     Token.Token_Id = FLOAT_FUNCT_TOKEN;
   }
   else
   {
     if (Token.Token_Id < VECTOR_FUNCT_TOKEN)
     {
       Token.Token_Id = VECTOR_FUNCT_TOKEN;
     }
     else
     {
       if (Token.Token_Id < COLOUR_KEY_TOKEN)
       {
         Token.Token_Id = COLOUR_KEY_TOKEN;
       }
     }
   }
}


/*****************************************************************************
*
* FUNCTION
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
* DESCRIPTION
*
* CHANGES
*
******************************************************************************/

static int Echo_getc()
{
  register int c;

  if (Got_EOF)
  {
    return(EOF);
  }
  
#ifdef FastMacroPatch
  c = Data_File_Getc (Data_File);
/*YS 03/2000 
  ignore '\r', this way Macintosh files can be read on all platforms */
  if ( c=='\r')
  {
    int c2=Data_File_Getc (Data_File);
    if ( c2 != '\n')
    {
      c='\n';
      Data_File_Ungetc (Data_File, c2);
    }
    else
      c=c2;
  }
/*YS 03/2000 */
  
#else
  c = getc(Data_File->File);
#endif
  
  if (c == EOF)
  {
#ifdef FastMacroPatch
    Data_File_Ungetc (Data_File, c);
#else
    ungetc(c,Data_File->File);
#endif
    c = '\n';
    Got_EOF=TRUE;
  }

  Echo_Ptr[Echo_Indx++] = c;

  if ((Echo_Indx > Echo_Line_Length) || (c == '\n'))
  {
    Echo_Ptr[Echo_Indx] = '\0';

    Echo_Indx = 0;

    Echo_Line++;

    if (Echo_Line == Num_Echo_Lines)
      Echo_Line = 0;

    Echo_Ptr=Echo_Buff[Echo_Line];
  }

  return(c);
}



/*****************************************************************************
*
* FUNCTION
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
* DESCRIPTION
*
* CHANGES
*
******************************************************************************/

#ifdef FastMacroPatch
void Echo_ungetc(int c)
#else
int Echo_ungetc(int c)
#endif
{
  if (Echo_Indx > 0)
  {
    Echo_Indx--;
  }

#ifdef FastMacroPatch
  Data_File_Ungetc (Data_File, c);
#else
  return(ungetc(c,Data_File->File));
#endif
}



/*****************************************************************************
*
* FUNCTION
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
* DESCRIPTION
*
* CHANGES
*
******************************************************************************/

void Where_Error ()
{
  int i;

  /* Return if no filename is specified. [DB 8/94] */

  if (Token.Filename == NULL)
  {
    return;
  }

  strcpy (&(Echo_Ptr[Echo_Indx])," <----ERROR\n");

  for (i=0;i<Num_Echo_Lines;i++)
  {
    Echo_Line++;

    if (Echo_Line==Num_Echo_Lines)
    {
      Echo_Line=0;
    }

    /* NK fix crash caused by '%' character in the error line */
    Error_Line_No_Param(Echo_Buff[Echo_Line]);
  }

  Error_Line("\n%s:%d: error: ", Token.Filename, Token.Token_Line_No+1);
}




/*****************************************************************************
*
* FUNCTION    Parse_Directive
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR      Chris Young
*
* DESCRIPTION
*
* CHANGES
*
******************************************************************************/

void Parse_Directive(int After_Hash)
{
  DBL Value, Value2;
  int Flag;
  char *ts;
  POV_MACRO *PMac=NULL;
  COND_TYPE Curr_Type = Cond_Stack[CS_Index].Cond_Type;
#ifdef FastMacroPatch
  long Hash_Loc = Data_File_Tell (Data_File);
#else
  long Hash_Loc = ftell(Data_File->File);
#endif
  
  if (Curr_Type == INVOKING_MACRO_COND)
  {
     if (Cond_Stack[CS_Index].PMac->Macro_End==Hash_Loc)
     {
        Return_From_Macro();
        if (--CS_Index < 0)
        {
           Error("Mis-matched '#end'.");
        }
        Token.Token_Id = END_OF_FILE_TOKEN;
        
        return;
     }
  }

  if (!Ok_To_Declare)
  {
    if (After_Hash)
    {
       Token.Token_Id=HASH_TOKEN; 
    }
    Token.Unget_Token = FALSE;

    return;
  }

  EXPECT
    CASE(IFDEF_TOKEN)
      Inc_CS_Index();

      if (Skipping)
      {
        Cond_Stack[CS_Index].Cond_Type = SKIP_TIL_END_COND;
        Skip_Tokens(SKIP_TIL_END_COND);
      }
      else
      {
        if (Parse_Ifdef_Param())
        {
           Cond_Stack[CS_Index].Cond_Type=IF_TRUE_COND;
        }
        else
        {
           Cond_Stack[CS_Index].Cond_Type=IF_FALSE_COND;
           Skip_Tokens(IF_FALSE_COND);
        }
      }
      EXIT
    END_CASE
    
    CASE(IFNDEF_TOKEN)
      Inc_CS_Index();

      if (Skipping)
      {
        Cond_Stack[CS_Index].Cond_Type = SKIP_TIL_END_COND;
        Skip_Tokens(SKIP_TIL_END_COND);
      }
      else
      {
        if (Parse_Ifdef_Param())
        {
           Cond_Stack[CS_Index].Cond_Type=IF_FALSE_COND;
           Skip_Tokens(IF_FALSE_COND);
        }
        else
        {
           Cond_Stack[CS_Index].Cond_Type=IF_TRUE_COND;
        }
      }
      EXIT
    END_CASE
    
    CASE(IF_TOKEN)
      Inc_CS_Index();

      if (Skipping)
      {
        Cond_Stack[CS_Index].Cond_Type = SKIP_TIL_END_COND;
        Skip_Tokens(SKIP_TIL_END_COND);
      }
      else
      {
        Value=Parse_Cond_Param();
      
        if (fabs(Value)>EPSILON)
        {
           Cond_Stack[CS_Index].Cond_Type=IF_TRUE_COND;
        }
        else
        {
           Cond_Stack[CS_Index].Cond_Type=IF_FALSE_COND;
           Skip_Tokens(IF_FALSE_COND);
        }
      }
      EXIT
    END_CASE

    CASE(WHILE_TOKEN)
      Inc_CS_Index();

      if (Skipping)
      {
        Cond_Stack[CS_Index].Cond_Type = SKIP_TIL_END_COND;
        Skip_Tokens(SKIP_TIL_END_COND);
      }
      else
      {
        Cond_Stack[CS_Index].While_File_Name = POV_STRDUP(Data_File->Filename);
#ifdef FastMacroPatch
        Cond_Stack[CS_Index].Pos        = Data_File_Tell (Data_File);
#else
        Cond_Stack[CS_Index].Pos        = ftell(Data_File->File);
#endif
        Cond_Stack[CS_Index].Line_No    = Data_File->Line_Number;

        Value=Parse_Cond_Param();
       
        if (fabs(Value)>EPSILON)
        {
           Cond_Stack[CS_Index].Cond_Type = WHILE_COND;
        }
        else
        {
           Cond_Stack[CS_Index].Cond_Type = SKIP_TIL_END_COND;
           Skip_Tokens(SKIP_TIL_END_COND);
        }
      }
      EXIT
    END_CASE
    
    CASE(ELSE_TOKEN)
      switch (Curr_Type)
      {
         case IF_TRUE_COND:
           Cond_Stack[CS_Index].Cond_Type = SKIP_TIL_END_COND;
           Skip_Tokens(SKIP_TIL_END_COND);
           break;
           
         case IF_FALSE_COND:
           Cond_Stack[CS_Index].Cond_Type = ELSE_COND;
           Token.Token_Id=HASH_TOKEN; /*insures Skip_Token takes notice*/
           UNGET
           break;
         
         case CASE_TRUE_COND:
         case SKIP_TIL_END_COND:
           break;

         case CASE_FALSE_COND:
           Cond_Stack[CS_Index].Cond_Type = CASE_TRUE_COND;
           if (Skipping)
           {
              Token.Token_Id=HASH_TOKEN; /*insures Skip_Token takes notice*/
              UNGET
           }
           break;

         default:
           Error("Mis-matched '#else'.");
      }
      EXIT
    END_CASE

    CASE(SWITCH_TOKEN)
      Inc_CS_Index();

      if (Skipping)
      {
        Cond_Stack[CS_Index].Cond_Type = SKIP_TIL_END_COND;
        Skip_Tokens(SKIP_TIL_END_COND);
      }
      else
      {
        Cond_Stack[CS_Index].Switch_Value=Parse_Cond_Param();
        Cond_Stack[CS_Index].Cond_Type=SWITCH_COND;
        EXPECT
          CASE2(CASE_TOKEN,RANGE_TOKEN)
            if (Token.Token_Id==CASE_TOKEN)
            {
              Value=Parse_Cond_Param();
              Flag = (fabs(Value-Cond_Stack[CS_Index].Switch_Value)<EPSILON);
            }
            else
            {
              Parse_Cond_Param2(&Value,&Value2);
              Flag = ((Cond_Stack[CS_Index].Switch_Value >= Value) &&
                      (Cond_Stack[CS_Index].Switch_Value <= Value2));
            }
 
            if(Flag)
            {
              Cond_Stack[CS_Index].Cond_Type=CASE_TRUE_COND;
            }
            else
            {
              Cond_Stack[CS_Index].Cond_Type=CASE_FALSE_COND;
              Skip_Tokens(CASE_FALSE_COND);
            }
            EXIT
          END_CASE

          OTHERWISE
            Error("#switch not followed by #case or #range.");
          END_CASE
        END_EXPECT
      }
      EXIT
    END_CASE

    CASE(BREAK_TOKEN)
      if (Curr_Type==CASE_TRUE_COND)
      {
        Cond_Stack[CS_Index].Cond_Type=SKIP_TIL_END_COND;
        Skip_Tokens(SKIP_TIL_END_COND);
      }          
      EXIT
    END_CASE
    
    CASE2(CASE_TOKEN,RANGE_TOKEN)
      switch(Curr_Type)
      {
        case CASE_TRUE_COND:
        case CASE_FALSE_COND:
          if (Token.Token_Id==CASE_TOKEN)
          {
            Value=Parse_Cond_Param();
            Flag = (fabs(Value-Cond_Stack[CS_Index].Switch_Value)<EPSILON);
          }
          else
          {
            Parse_Cond_Param2(&Value,&Value2);
            Flag = ((Cond_Stack[CS_Index].Switch_Value >= Value) &&
                    (Cond_Stack[CS_Index].Switch_Value <= Value2));
          }
          
          if(Flag && (Curr_Type==CASE_FALSE_COND))
          {
            Cond_Stack[CS_Index].Cond_Type=CASE_TRUE_COND;
            if (Skipping)
            {
               Token.Token_Id=HASH_TOKEN; /*insures Skip_Token takes notice*/
               UNGET
            }
          }
          break;           

        case SWITCH_COND:
          UNGET
        case SKIP_TIL_END_COND:
          break;

        default:
          Error("Mis-matched '#case' or '#range'.");
      }
      EXIT
    END_CASE
    
    CASE(END_TOKEN)
      switch (Curr_Type)
      {
         case INVOKING_MACRO_COND:
           Return_From_Macro();
           if (--CS_Index < 0)
           {
              Error("Mis-matched '#end'.");
           }
           break;
         
         case IF_FALSE_COND:
           Token.Token_Id=HASH_TOKEN; /*insures Skip_Token takes notice*/
           UNGET
         case IF_TRUE_COND:
         case ELSE_COND:
         case CASE_TRUE_COND:
         case CASE_FALSE_COND:
         case DECLARING_MACRO_COND:
         case SKIP_TIL_END_COND:
           if (Curr_Type==DECLARING_MACRO_COND)
           {
             if ((PMac=Cond_Stack[CS_Index].PMac)!=NULL)
             {
#ifdef FastMacroPatch
                PMac->Text = NULL;
#endif
                PMac->Macro_End=Hash_Loc;
             }
           }
           if (--CS_Index < 0)
           {
              Error("Mis-matched '#end'.");
           }
           if (Skipping)
           {
              Token.Token_Id=HASH_TOKEN; /*insures Skip_Token takes notice*/
              UNGET
           }
           break;
         
         case WHILE_COND:
           if (strcmp(Cond_Stack[CS_Index].While_File_Name,Data_File->Filename))
           { 
              Error("#while loop didn't end in file where it started.");
           }
           
           Got_EOF=FALSE;
#ifdef FastMacroPatch
           if (Data_File_Seek (Data_File, Cond_Stack[CS_Index].Pos) < 0)
#else
           if (fseek(Data_File->File, Cond_Stack[CS_Index].Pos,0) < 0)
#endif
           {
              Error("Unable to seek in input file for #while directive.\n");
           }

           Data_File->Line_Number = Cond_Stack[CS_Index].Line_No;

           Value=Parse_Cond_Param();
      
           if (fabs(Value)<EPSILON)
           {
             POV_FREE(Cond_Stack[CS_Index].While_File_Name);
             Cond_Stack[CS_Index].While_File_Name=NULL;
             Cond_Stack[CS_Index].Cond_Type = SKIP_TIL_END_COND;
             Skip_Tokens(SKIP_TIL_END_COND);
           }
           break;

         default:
           Error("Mis-matched '#end'.");
      }
      EXIT
    END_CASE

  	  #ifdef SetPatch
  	  	CASE(SET_TOKEN)
  	  		#ifdef UnofficialBlocking
      			parseUnofficialFeature(70);
				#endif	
			#endif
  	  	CASE2(DECLARE_TOKEN, LOCAL_TOKEN)
      if (Skipping)
      {
         UNGET
         EXIT
      }
      else
      {
		#ifdef SetPatch
         Parse_Declare(FALSE);
		#else
         Parse_Declare ();
		#endif        
		 Curr_Type = Cond_Stack[CS_Index].Cond_Type;
         if (Token.Unget_Token)
         {
           switch (Token.Token_Id)
           {
              case HASH_TOKEN:
                Token.Unget_Token=FALSE;
                break;
              
              case MACRO_ID_TOKEN:
                break;
              
              default:
                EXIT
           }
         }
         else
         {
            EXIT
         }
      }
    END_CASE
    
    CASE (DEFAULT_TOKEN)
      Parse_Default();
      EXIT
    END_CASE

  #ifdef BsplinePatch

    CASE (INIT_SPLINE_TOKEN)
      Parse_Init_Spline();
      EXIT
    END_CASE

    CASE (INIT_3D_SPLINE_TOKEN)
      Parse_Init_3D_Spline();
      EXIT
    END_CASE
  #endif

    CASE (INCLUDE_TOKEN)
      if (Skipping)
      {
         UNGET
      }
      else
      {
         Open_Include();
      }
      EXIT
    END_CASE

    CASE (FLOAT_FUNCT_TOKEN)
      if (Skipping)
      {
        UNGET
        EXIT
      }
      else
      {
        switch(Token.Function_Id)
        {
           case VERSION_TOKEN:
             Ok_To_Declare = FALSE;
#ifdef UnofficialBlocking
             {
               DBL newVersion;

               /* try to read a float */
               newVersion = Allow_Float(-100.0);

               if (newVersion < -99.0)
               {
                 /* -------- string token was read --------- */
                 Get_Token();

                 if (!pov_stricmp(Token.Token_String,"official"))
                 {
                   /* -------- official --------- */
                   /* 
                      file was designed for the official version only
                      we now disable unofficial-version features
                   */
                   opts.Language_Version = (int)(Parse_Float()*100.0+0.5);
                   Data_File->unofficialVersion = -1;

                   if (UNOFFICIAL_VERSION_STRING != NULL)
                   {
                     /* this is the official version */
                     Warning(0,"Setting language version to %g.  Disabling unofficial features.\n", newVersion);
                   }
                 }
                 else if(!pov_stricmp(Token.Token_String,"unofficial"))
                 {
                   /* --------- unofficial ---------- */
                   if (UNOFFICIAL_VERSION_STRING == NULL)
                   {
                     /* -------- unofficial BUT THIS IS THE OFFICIAL --------- */
                     /* we are using the official version, but the file
                        was designed for an unofficial version
                     */
                     Error("This file was designed for %s, which is an unofficial compile of POV-Ray.\nIt may not work with this official version.\n",
                       Token.Token_String);
                   }

                   /* get the name of the unofficial version */
                   Get_Token();

                   if(!pov_stricmp(Token.Token_String,UNOFFICIAL_VERSION_STRING))
                   {
                     /* -------- unofficial - MATCH --------- */
                     /* we have a match */
                     /* read version number */
                     Data_File->unofficialVersion = (int)(Parse_Float()*100.0+0.5);

                     if (Data_File->unofficialVersion < UNOFFICIAL_VERSION_NUMBER)
                     {
                       /* older file with newer version */
                       /* what to do? */
                       /* let's just give a warning */
                       Warning(0, "This file was designed for (unofficial) %s version %g.  You are using version %g.\n",
                         UNOFFICIAL_VERSION_STRING,Data_File->unofficialVersion/100.0, UNOFFICIAL_VERSION_NUMBER/100.0);
                       Warning(0, "Some features may have changed.\n");
                     }
                     else if (Data_File->unofficialVersion > UNOFFICIAL_VERSION_NUMBER)
                     {
                       /* file is too new for this version - give an error */
                       Error("This file was designed for (unofficial) %s version %g.  You are using version %g.\n",
                         UNOFFICIAL_VERSION_STRING,Data_File->unofficialVersion/100.0, UNOFFICIAL_VERSION_NUMBER/100.0);
                     }
                   }
                   else
                   {
                     /* -------- unofficial - NO MATCH --------- */
                     /* non-matching unofficial string */
                     Error("This file was designed for (unofficial) %s.  You are using %s instead.\n",
                       Token.Token_String, UNOFFICIAL_VERSION_STRING);
                   }
                 }
                 else if (!pov_stricmp(Token.Token_String,"development"))
                 {
                   /* -------- development --------- */
                   /* 
                      file was designed for the development version
                   */
#if OFFICIAL_DEVELOPMENT_VERSION
                   opts.Language_Version = (int)(Parse_Float()*100.0+0.5);
                   if (opts.Language_Version > OFFICIAL_VERSION_NUMBER)
                   {
                     Error("This file requires POV-Ray development version %g.\n",opts.Language_Version/100.0);
                   }
#else
                   Error("This file requires POV-Ray development version %g.\n",opts.Language_Version/100.0);
#endif
                 }
                 else
                 {
                   Error("Must be version number, 'unofficial', 'official', or 'development'.\n");
                 }
               }
               else
               {
                 /* -------- Just a version number was found --------- */
                 /* simply set the official version number */
                 opts.Language_Version = (int)(newVersion*100.0+0.5);

                 /* for megapov */
                 if (opts.Language_Version > 310 )
                 {
                   Data_File->unofficialVersion = 30;
                 }
                 else if (opts.Language_Version > OFFICIAL_VERSION_NUMBER)
                 {
                   Error("This file requires POV-Ray version %g.\n",opts.Language_Version/100.0);
                 }
               }

             }
             opts.unofficialVersion = Data_File->unofficialVersion;
#else
             opts.Language_Version = (int)(Parse_Float ()*100.0+0.5);
#endif
             Parse_Semi_Colon();
             Ok_To_Declare = TRUE;
             Curr_Type = Cond_Stack[CS_Index].Cond_Type;
             if (Token.Unget_Token && (Token.Token_Id==HASH_TOKEN))
             {
                Token.Unget_Token=FALSE;
             }
             else
             {
                EXIT
             }
             break;

           default:
             UNGET
             Parse_Error_Str ("object or directive.");
             break;
        }
      }
    END_CASE

    CASE(WARNING_TOKEN)
      if (Skipping)
      {
        UNGET
      }
      else
      {     
        ts=Parse_Formatted_String();
        Warning(0,ts);
        POV_FREE(ts);
      }
      EXIT
    END_CASE
      
    CASE(ERROR_TOKEN)
      if (Skipping)
      {
        UNGET
      }
      else
      {     
        ts=Parse_Formatted_String();
        POV_FREE(ts);
        Error("User error directive hit.");
      }
      EXIT
    END_CASE
      
    CASE(RENDER_TOKEN)
      if (Skipping)
      {
        UNGET
      }
      else
      {     
        ts=Parse_Formatted_String();
        Render_Info(ts);
        POV_FREE(ts);
      }
      EXIT
    END_CASE
      
    CASE(STATISTICS_TOKEN)
      if (Skipping)
      {
        UNGET
      }
      else
      {     
        ts=Parse_Formatted_String();
        Statistics(ts);
        POV_FREE(ts);
      }
      EXIT
    END_CASE
      
    CASE(DEBUG_TOKEN)
      if (Skipping)
      {
        UNGET
      }
      else
      {     
        ts=Parse_Formatted_String();
        Debug_Info(ts);
        POV_FREE(ts);
      }
      EXIT
    END_CASE

    CASE(FOPEN_TOKEN)
      if (Skipping)
      {
        UNGET
      }
      else
      {
        Parse_Fopen();
      }
      EXIT
    END_CASE

    CASE(FCLOSE_TOKEN)
      if (Skipping)
      {
        UNGET
      }
      else
      {
        Parse_Fclose();
      }
      EXIT
    END_CASE

    CASE(READ_TOKEN)
      if (Skipping)
      {
        UNGET
      }
      else
      {
        Parse_Read();
      }
      EXIT
    END_CASE

    CASE(WRITE_TOKEN)
      if (Skipping)
      {
        UNGET
      }
      else
      {
        Parse_Write();
      }
      EXIT
    END_CASE

    CASE(UNDEF_TOKEN)
      if (Skipping)
      {
        UNGET
      }
      else
      {
        EXPECT
          CASE (IDENTIFIER_TOKEN)
            Warn(0,"Attempt to undef unknown identifier");
            EXIT
          END_CASE

          CASE4 (TNORMAL_ID_TOKEN, FINISH_ID_TOKEN, TEXTURE_ID_TOKEN, OBJECT_ID_TOKEN)
          CASE4 (COLOUR_MAP_ID_TOKEN, TRANSFORM_ID_TOKEN, CAMERA_ID_TOKEN, PIGMENT_ID_TOKEN)
          CASE4 (SLOPE_MAP_ID_TOKEN,NORMAL_MAP_ID_TOKEN,TEXTURE_MAP_ID_TOKEN,COLOUR_ID_TOKEN)
          CASE4 (PIGMENT_MAP_ID_TOKEN, MEDIA_ID_TOKEN,STRING_ID_TOKEN,INTERIOR_ID_TOKEN)
          CASE4 (ARRAY_ID_TOKEN, DENSITY_ID_TOKEN, DENSITY_MAP_ID_TOKEN, FILE_ID_TOKEN)
          CASE4 (FOG_ID_TOKEN, MACRO_ID_TOKEN, PARAMETER_ID_TOKEN, RAINBOW_ID_TOKEN)
          CASE4 (SKYSPHERE_ID_TOKEN,MATERIAL_ID_TOKEN,UV_ID_TOKEN,VECTOR_4D_ID_TOKEN)
            Remove_Symbol (Token.Table_Index,Token.Token_String);
            EXIT
          END_CASE
        
          CASE2 (VECTOR_FUNCT_TOKEN, FLOAT_FUNCT_TOKEN)
            switch(Token.Function_Id)
            {
              case VECTOR_ID_TOKEN:
              case FLOAT_ID_TOKEN:
                 Remove_Symbol (Token.Table_Index,Token.Token_String);
                 break;

              default:
                 Parse_Error(IDENTIFIER_TOKEN);
                 break;
            }
            EXIT
          END_CASE

          OTHERWISE
            Parse_Error(IDENTIFIER_TOKEN);
          END_CASE
        END_EXPECT
      }
      EXIT
    END_CASE

    CASE (MACRO_ID_TOKEN)
      if (Skipping)
      {
        UNGET
      }
      else
      { 
        Invoke_Macro();
      }
      EXIT
    END_CASE

    /* begin added djconnel : 12 Sep 1998 */
/*
    CASE(SYSTEM_TOKEN)
      if (Skipping)
      {
        UNGET
      }     
	    else
      {
	      ts=Parse_Formatted_String();
	      system(ts);
	      POV_FREE(ts);
      }
      EXIT
    END_CASE
*/

    CASE (MACRO_TOKEN)
      if (!Skipping)
      {
        if (Inside_MacroDef)
        {
          Error("Cannot nest macro definitions");
        }
        Inside_MacroDef=TRUE;
        PMac=Parse_Macro();
        Inside_MacroDef=FALSE;
      }
      Inc_CS_Index();
      Cond_Stack[CS_Index].Cond_Type = DECLARING_MACRO_COND;
      Cond_Stack[CS_Index].PMac      = PMac;
      Skip_Tokens(DECLARING_MACRO_COND);
      EXIT
    END_CASE

#ifdef CreateLabelPatch
    /* NK persist create Nov 1999 */
    CASE (CREATE_TOKEN)
      Get_Token();

      if (createObjectName!=NULL)
        POV_FREE(createObjectName);
      createObjectName = POV_STRDUP(Token.Token_String);

      GET (EQUALS_TOKEN);

      EXIT
    END_CASE
    /* NK ---- */
#endif

    OTHERWISE
      UNGET
      EXIT
    END_CASE
  END_EXPECT
  
  if (Token.Unget_Token)
  {
    Token.Unget_Token = FALSE;
  }
  else
  {
    Token.Token_Id = END_OF_FILE_TOKEN;
  }
}

/* NK 1999 - use $ for #declare */
/*****************************************************************************
*
* FUNCTION    Parse_Assignment()
*
* AUTHOR      Nathan Kopp
*
* DESCRIPTION 
*     Parses an assignment.  This previously was part of Parse_Directive()
*     as it related to the "#declare" directive.  The new syntax allows
*     declarations in the form:   $a=b; instead of #declare a=b;
*
*     This is now called by both Parse_Directive() (for #declare) and
*     Get_Token()
*
* CHANGES
*
******************************************************************************/
static int Parse_Assignment()
{
  int finished=FALSE;

  parseUnofficialFeature(10);

  while(!finished)
  {
    finished=TRUE;

    /* this skipping stuff might not work!!! */
    if (Skipping)
    {
       UNGET
    }
    else
    {
		#ifdef SetPatch
         Parse_Declare(FALSE);	
		#else
         Parse_Declare ();
		#endif
       if (Token.Unget_Token)
       {
         switch (Token.Token_Id)
         {
            case HASH_TOKEN:
              Token.Unget_Token=FALSE;
              Parse_Directive(TRUE);
              break;
            
            case MACRO_ID_TOKEN:
              Parse_Directive(FALSE);
              break;

            case DOLLAR_TOKEN:
            case PERCENT_TOKEN:
              finished=FALSE;
              break;
            
            default:
              break;
         }
       }
    }

    if (Token.Unget_Token)
    {
      Token.Unget_Token = FALSE;
    }
    else
    {
      Token.Token_Id = END_OF_FILE_TOKEN;
    }
  }
  return FALSE;
}
/* NK ---- */

/*****************************************************************************
*
* FUNCTION    Get_Include_File_Depth
*
* INPUT       -none-
*
* OUTPUT      -none-
*
* RETURNS     The index of the currently opened file (0, 1, 2...)
*
* AUTHOR      Eduard Schwan
*
* DESCRIPTION Call this from GUI to determine which file was being parsed
*             When a fatal error happened.  The index can be used to access
*             the file array (retrieved with the Get_File_Array() call.)
*
* CHANGES
*
******************************************************************************/

int Get_Include_File_Depth(void)
{
	return Include_File_Index;
}


/*****************************************************************************
*
* FUNCTION    Get_Include_File_Array
*
* INPUT       -none-
*
* OUTPUT      -none-
*
* RETURNS     A pointer to the array of file records.
*
* AUTHOR      Eduard Schwan
*
* DESCRIPTION Call this from GUI to determine which file was being parsed
*             When a fatal error happened.
*
* CHANGES
*
******************************************************************************/

DATA_FILE * Get_Include_File_Array(void)
{
	return Include_Files;
}


/*****************************************************************************
*
* FUNCTION
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
* DESCRIPTION
*
* CHANGES
*
******************************************************************************/

void Open_Include()
{
   char *temp;
   char b[FILE_NAME_LENGTH];

   if (Skip_Spaces () != TRUE)
     Error ("Expecting a string after INCLUDE.\n");

   Include_File_Index++;

   if (Include_File_Index >= MAX_INCLUDE_FILES)
   {
     Include_File_Index--;
     Error ("Too many nested include files.\n");
   }

   temp = Parse_String();

   UICB_OPEN_INCLUDE_FILE  /* Notify UI that we are about to open an include file */

   Echo_Ptr[Echo_Indx++] = '\n';
   Echo_Ptr[Echo_Indx] = '\0';
   Echo_Indx = 0;
   Echo_Line++;
   if (Echo_Line == Num_Echo_Lines)
   {
     Echo_Line = 0;
   }
   Echo_Ptr=Echo_Buff[Echo_Line];

   Data_File = &Include_Files[Include_File_Index];
   Data_File->Line_Number = 0;

#ifdef FastMacroPatch
   Fopen_Setup (Data_File);
   if ((Data_File->Data = Locate_File (temp, READ_BINFILE_STRING,".inc",".INC",b,TRUE)) == NULL)
#else
   if ((Data_File->File = Locate_File (temp, READ_TXTFILE_STRING,".inc",".INC",b,TRUE)) == NULL)
#endif
   {
      Data_File->Filename = NULL;  /* Keeps from closing failed file. */
      Stage=STAGE_INCLUDE_ERR;
      Error ("Cannot open include file %s.\n", temp);
   }
  ADDINCLUDEMARKER
   
   POV_FREE(temp);

   Data_File->Filename = POV_STRDUP(b);
   Data_File->R_Flag=FALSE;
#ifdef UnofficialBlocking
   Data_File->unofficialVersion = -1;
   opts.unofficialVersion = Data_File->unofficialVersion;
#endif

   Add_Sym_Table(Data_File->Filename);

   Token.Token_Id = END_OF_FILE_TOKEN;
   
}



/*****************************************************************************
*
* FUNCTION
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
* DESCRIPTION
*
* CHANGES
*
******************************************************************************/

static void Skip_Tokens(COND_TYPE cond)
{
  int Temp      = CS_Index;
  int Prev_Skip = Skipping;

  Skipping=TRUE;

  while ((CS_Index > Temp) || ((CS_Index == Temp) && (Cond_Stack[CS_Index].Cond_Type == cond)))
  {
    Get_Token();
  }

  Skipping=Prev_Skip;

  if (Token.Token_Id==HASH_TOKEN)
  {
     Token.Token_Id=END_OF_FILE_TOKEN;
     Token.Unget_Token=FALSE;
  }
  else
  {
     UNGET
  }
}


/*****************************************************************************
*
* FUNCTION
*
*   get_hash_value
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   Dieter Bayer
*
* DESCRIPTION
*
*   Calculate hash value for a given string.
*
* CHANGES
*
*   Apr 1996 : Creation.
*
******************************************************************************/

static int get_hash_value(char *s)
{
  unsigned int i = 0;

  while (*s)
  {
    i = (i << 1) ^ *s++;
  }

  return((int)(i % SYM_TABLE_SIZE));
}



/*****************************************************************************
*
* FUNCTION
*
*   init_sym_tables
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   Chris Young
*
* DESCRIPTION
*
* CHANGES
*
******************************************************************************/

static void init_sym_tables()
{
  int i;

  Add_Sym_Table("reserved words");

  for (i = 0; i < LAST_TOKEN; i++)
  {
    Add_Symbol(0,Reserved_Words[i].Token_Name,Reserved_Words[i].Token_Number);
  }

  Add_Sym_Table("global identifiers");
}

static void Add_Sym_Table(char *s)
{
  int i;

  SYM_TABLE *New;

  if ((++Table_Index)==MAX_NUMBER_OF_TABLES)
  {
    Table_Index--;
    Error("Too many nested symbol tables");
  }
  
  Tables[Table_Index]=New=(SYM_TABLE *)POV_MALLOC(sizeof(SYM_TABLE),"symbol table");
  
  New->Table_Name=POV_STRDUP(s);

  for (i = 0; i < SYM_TABLE_SIZE; i++)
  {
    New->Table[i] = NULL;
  }

}

static void Destroy_Table(int index)
{
   int i;
   SYM_TABLE *Table = Tables[index];
   SYM_ENTRY *Entry;
   
   for (i = 0 ; i < SYM_TABLE_SIZE; i++)
   {
      Entry=Table->Table[i];

      while (Entry)
      {
         Entry = Destroy_Entry(Entry);
      }
   }

   POV_FREE(Table->Table_Name);
   POV_FREE(Table);
   
}


SYM_ENTRY *Create_Entry (int Index,char *Name,TOKEN Number)
{
  SYM_ENTRY *New;
  
  New = (SYM_ENTRY *)POV_MALLOC(sizeof(SYM_ENTRY), "symbol table entry");

  New->Token_Number = Number;
  New->Data         = NULL;
  if (Index)
  {
     New->Token_Name= POV_STRDUP(Name);
  }
  else
  {
     New->Token_Name=Name;
  }
  
  return(New);
}

SYM_ENTRY *Destroy_Entry (SYM_ENTRY *Entry)
{
   SYM_ENTRY *Next;
   
   if (Entry==NULL)
   {
      return(NULL);
   }

   Next=Entry->next;

   /* if data is NULL then this is a reserved word and the name
      doesn't need freed.
    */
        
   if (Entry->Data)
   {
     POV_FREE(Entry->Token_Name);
   }

   Destroy_Ident_Data (Entry->Data,Entry->Token_Number);
    
   POV_FREE(Entry);

   return(Next);
}


static void Add_Entry (int Index,SYM_ENTRY *Table_Entry)
{
  int i = get_hash_value(Table_Entry->Token_Name);

  Table_Entry->next       = Tables[Index]->Table[i];
  Tables[Index]->Table[i] = Table_Entry;
}


SYM_ENTRY *Add_Symbol (int Index,char *Name,TOKEN Number)
{
  SYM_ENTRY *New;

  New = Create_Entry (Index,Name,Number);
  Add_Entry(Index,New);

  return(New);
}


#ifdef SetPatch
	SYM_ENTRY *Find_Symbol(int Index,char *Name)
#else
	static SYM_ENTRY *Find_Symbol(int Index,char *Name)
#endif
{
  SYM_ENTRY *Entry;
  
  int i = get_hash_value(Name);

  Entry = Tables[Index]->Table[i];

  while (Entry)
  {
    if (strcmp(Name, Entry->Token_Name) == 0)
    {
      return(Entry);
    }

    Entry = Entry->next;
  }

  return(Entry);
}


static void Remove_Symbol (int Index, char *Name)
{
  SYM_ENTRY *Entry;
  SYM_ENTRY **EntryPtr;
  
  int i = get_hash_value(Name);

  EntryPtr = &(Tables[Index]->Table[i]);
  Entry    = *EntryPtr;

  while (Entry)
  {
    if (strcmp(Name, Entry->Token_Name) == 0)
    {
      *EntryPtr = Entry->next;
      Destroy_Entry(Entry);
      return;
    }
    
    EntryPtr = &(Entry->next);
    Entry    = *EntryPtr;
  }
  
  Error("Tried to free undefined symbol");
}

static void Check_Macro_Vers(void)
{
  if (opts.Language_Version < 310)
  {
     Error("Macros require #version 3.1 but #version %4.2lf is set.\n",
            opts.Language_Version/100.0);
  }
}

static POV_MACRO *Parse_Macro()
{
  POV_MACRO *New;
  SYM_ENTRY *Table_Entry;
  int Old_Ok = Ok_To_Declare;

  Check_Macro_Vers();
  
  Ok_To_Declare = FALSE;

  EXPECT
    CASE (IDENTIFIER_TOKEN)
      Table_Entry = Add_Symbol (1,Token.Token_String,MACRO_ID_TOKEN);
      EXIT
    END_CASE

    CASE (MACRO_ID_TOKEN)
      Remove_Symbol(1,Token.Token_String);
      Table_Entry = Add_Symbol (1,Token.Token_String,MACRO_ID_TOKEN);
      EXIT
    END_CASE
    
    OTHERWISE
      Parse_Error(IDENTIFIER_TOKEN);
      Table_Entry = NULL; /* tw */
    END_CASE
  END_EXPECT
  
  Ok_To_Declare = Old_Ok;

  New=(POV_MACRO *)POV_MALLOC(sizeof(POV_MACRO),"macro");

  Table_Entry->Data=(void *)New;

  New->Macro_Filename = NULL;
  New->Macro_Name=POV_STRDUP(Token.Token_String);
#ifdef FastMacroPatch
  New->Text = NULL;
  New->Parent_Text = NULL;
#endif

#ifdef UnofficialBlocking
  New->unofficialVersion=Data_File->unofficialVersion;
#endif
  
  GET (LEFT_PAREN_TOKEN);
  
  New->Num_Of_Pars=0;
  
  EXPECT
    CASE2 (IDENTIFIER_TOKEN,PARAMETER_ID_TOKEN)
    CASE4 (TNORMAL_ID_TOKEN, FINISH_ID_TOKEN, TEXTURE_ID_TOKEN, OBJECT_ID_TOKEN)
    CASE4 (COLOUR_MAP_ID_TOKEN, TRANSFORM_ID_TOKEN, CAMERA_ID_TOKEN, PIGMENT_ID_TOKEN)
    CASE4 (SLOPE_MAP_ID_TOKEN,NORMAL_MAP_ID_TOKEN,TEXTURE_MAP_ID_TOKEN,COLOUR_ID_TOKEN)
    CASE4 (PIGMENT_MAP_ID_TOKEN, MEDIA_ID_TOKEN,STRING_ID_TOKEN,INTERIOR_ID_TOKEN)
    CASE4 (ARRAY_ID_TOKEN, DENSITY_ID_TOKEN, DENSITY_MAP_ID_TOKEN, FILE_ID_TOKEN)
    CASE4 (FOG_ID_TOKEN, RAINBOW_ID_TOKEN, SKYSPHERE_ID_TOKEN,MATERIAL_ID_TOKEN)
    CASE2 (UV_ID_TOKEN,VECTOR_4D_ID_TOKEN)
      New->Par_Name[New->Num_Of_Pars] = POV_STRDUP(Token.Token_String);
      if (++(New->Num_Of_Pars) == MAX_PARAMETER_LIST)
      {
        Error("Too many parameters");
      }
      Parse_Comma();
    END_CASE
    
    CASE2 (VECTOR_FUNCT_TOKEN, FLOAT_FUNCT_TOKEN)
      switch(Token.Function_Id)
      {
         case VECTOR_ID_TOKEN:
         case FLOAT_ID_TOKEN:
           New->Par_Name[New->Num_Of_Pars] = POV_STRDUP(Token.Token_String);
           if (++(New->Num_Of_Pars) == MAX_PARAMETER_LIST)
           {
             Error("Too many parameters");
           }
           Parse_Comma();
           break;

         default:
           Parse_Error_Str ("identifier or expression.");
           break;
      }
    END_CASE
    
    CASE(RIGHT_PAREN_TOKEN)
      UNGET
      EXIT
    END_CASE
    
    OTHERWISE
      Parse_Error_Str ("identifier or expression.");
    END_CASE
  END_EXPECT
  
  if (!strcmp("memory macro",Data_File->Filename))
  {
    /*oops - nested macro definitions are not allowed anymore*/
    /*Error("Nested #macro definitons are not allowed in MegaPov.\nPlease separate the definitions.\n");*/
    New->Parent_Text = ((char **) Data_File->Data) [1];
  }

  New->Macro_Filename = POV_STRDUP(Data_File->Filename);
#ifdef FastMacroPatch
  New->Macro_Pos      = Data_File_Tell (Data_File);
#else
  New->Macro_Pos      = ftell(Data_File->File);
#endif
  New->Macro_Line_No  = Data_File->Line_Number;

  Check_Macro_Vers();

  return (New);
}


static void Invoke_Macro()
{
  POV_MACRO *PMac=(POV_MACRO *)Token.Data;
  SYM_ENTRY **Table_Entries = NULL; /* tw */
  int i;
#ifndef FastMacroPatch
  char *f;
#endif
  
  Check_Macro_Vers();

  GET(LEFT_PAREN_TOKEN);
  
  if (PMac->Num_Of_Pars > 0)
  {
    Table_Entries = (SYM_ENTRY **)POV_MALLOC(sizeof(SYM_ENTRY *)*PMac->Num_Of_Pars,"parameters");

    /* We must parse all parameters before adding new symbol table
       or adding entries.  Otherwise recursion won't always work.
     */

    for (i=0; i<PMac->Num_Of_Pars; i++)
    {
      Table_Entries[i]=Create_Entry(1,PMac->Par_Name[i],IDENTIFIER_TOKEN);
      if (!Parse_RValue(IDENTIFIER_TOKEN,&(Table_Entries[i]->Token_Number),&(Table_Entries[i]->Data),TRUE,FALSE))
      {
        Error("Expected %d parameters but only %d found.",PMac->Num_Of_Pars,i);
      }
      Parse_Comma();
    }
  }
  
  GET(RIGHT_PAREN_TOKEN);
  
  Inc_CS_Index();
  Cond_Stack[CS_Index].Cond_Type = INVOKING_MACRO_COND;
  
  Cond_Stack[CS_Index].Pos               = Data_File_Tell(Data_File);
  Cond_Stack[CS_Index].Line_No           = Data_File->Line_Number;
  Cond_Stack[CS_Index].Macro_Return_Name = Data_File->Filename;
  Cond_Stack[CS_Index].PMac              = PMac;
#ifdef UnofficialBlocking
  Cond_Stack[CS_Index].unofficialVersion = Data_File->unofficialVersion;
#endif
  
  /* Gotta have new symbol table in case #local is used */
  Add_Sym_Table(PMac->Macro_Name);
  
  if (PMac->Num_Of_Pars > 0)
  {
    for (i=0; i<PMac->Num_Of_Pars; i++)
    {
      Add_Entry(Table_Index,Table_Entries[i]);
    }

    POV_FREE(Table_Entries);
  }

#ifdef FastMacroPatch
  if (PMac->Text==NULL)
  {
    unsigned int Length = PMac->Macro_End - PMac->Macro_Pos + 3;

    PMac->Text =(char*) POV_MALLOC (Length + 1, "macro text");

    if (PMac->Parent_Text != NULL)
    {
      /* copy macro from string - nested macro */
      unsigned int i;

      for(i=0; i<Length; i++)
      {
        PMac->Text[i] = PMac->Parent_Text[i+PMac->Macro_Pos];
      }

      PMac->Text[i] = 0;
    }
    else if (strcmp(PMac->Macro_Filename,Data_File->Filename))
    {
      /* Not in same file */
      int z;
      DATA_FILE Data_File2;
      Data_File2.Filename = POV_STRDUP(PMac->Macro_Filename);
      Data_File2.R_Flag=FALSE;

      Fopen_Setup((&Data_File2));
      if ((Data_File2.Data = Locate_File (Data_File2.Filename, READ_BINFILE_STRING,"","",NULL,TRUE)) == NULL)
      {
         char *f = Data_File2.Filename;
         Data_File2.Filename = NULL;  /* Keeps from closing failed file. */
         Stage=STAGE_INCLUDE_ERR;
         Error ("Cannot open macro file %s.\n", f);
      }

      if (Data_File_Seek ((&Data_File2), PMac->Macro_Pos) < 0)
         Error ("Unable to seek to beginning of macro in #macro");

      if ((z=fread (PMac->Text, 1, Length, (FILE*)Data_File2.Data)) != Length)
         Error ("Unable to read #macro");

      Data_File_Close((&Data_File2));
    }
    else
    {
      int savepos = Data_File_Tell(Data_File);

      if (Data_File_Seek (Data_File, PMac->Macro_Pos) < 0)
         Error ("Unable to seek to beginning of macro in #macro");

      /* Not portable, but reasonable here */
      if (fread (PMac->Text, 1, Length, (FILE*)Data_File->Data) != Length)
         Error ("Unable to read #macro");
     
      if (Data_File_Seek (Data_File, savepos) < 0)
         Error ("Unable to seek to end of macro after #macro");
    }

    PMac->Macro_End = PMac->Macro_End - PMac->Macro_Pos;

    PMac->Text [Length] = '\0';
  }


  Cond_Stack[CS_Index].Return = *Data_File;
  Data_File->Line_Number = PMac->Macro_Line_No;
  Data_File->Data = POV_MALLOC (sizeof (char *) * 2, "file data");
  ((char **) Data_File->Data) [0] = PMac->Text;
  ((char **) Data_File->Data) [1] = PMac->Text;
  Data_File->Close = String_Close;
  Data_File->Getc = String_Getc;
  Data_File->Ungetc = String_Ungetc;
  Data_File->Putc = String_Putc;
  Data_File->Tell = String_Tell;
  Data_File->Seek = String_Seek;
  Data_File->Filename = POV_STRDUP("memory macro");/*POV_STRDUP (PMac->Macro_Filename);*/
    
  Got_EOF=FALSE;

#else
  if (strcmp(PMac->Macro_Filename,Data_File->Filename))
  {
    /* Not in same file */
    Cond_Stack[CS_Index].Macro_Same_Flag=FALSE;
    fclose(Data_File->File);
    Got_EOF=FALSE;
    Data_File->Filename = POV_STRDUP(PMac->Macro_Filename);
    Data_File->R_Flag=FALSE;

#ifdef UnofficialBlocking
    Data_File->unofficialVersion=PMac->unofficialVersion;
    opts.unofficialVersion = Data_File->unofficialVersion;
#endif

    if ((Data_File->File = Locate_File (Data_File->Filename, READ_TXTFILE_STRING,"","",NULL,TRUE)) == NULL)
    {
       f = Data_File->Filename;
       Data_File->Filename = NULL;  /* Keeps from closing failed file. */
       Stage=STAGE_INCLUDE_ERR;
       Error ("Cannot open macro file %s.\n", f);
    }
  }
  else
  {
    Cond_Stack[CS_Index].Macro_Same_Flag=TRUE;
  }

  Data_File->Line_Number=PMac->Macro_Line_No;

  Got_EOF=FALSE;
  if (fseek(Data_File->File, PMac->Macro_Pos,0) < 0)
  {
    Error("Unable to file seek in macro.\n");
  }
#endif

  Token.Token_Id = END_OF_FILE_TOKEN;

  Check_Macro_Vers();

}

static void Return_From_Macro()
{
#ifndef FastMacroPatch
  char *f;
#endif

  Check_Macro_Vers();

#ifdef FastMacroPatch
  POV_FREE (Data_File->Data); /* should do a 'close_file' instead */
  POV_FREE (Data_File->Filename);
  *Data_File = Cond_Stack[CS_Index].Return;
#ifdef UnofficialBlocking
  opts.unofficialVersion = Data_File->unofficialVersion;
#endif

#else
  if (!Cond_Stack[CS_Index].Macro_Same_Flag)
  {
     fclose(Data_File->File);
     Got_EOF=FALSE;
     POV_FREE(Data_File->Filename);
     Data_File->Filename = Cond_Stack[CS_Index].Macro_Return_Name;
     Data_File->R_Flag=FALSE;

#ifdef UnofficialBlocking
     Data_File->unofficialVersion = Cond_Stack[CS_Index].unofficialVersion;
     opts.unofficialVersion = Data_File->unofficialVersion;
#endif

     if ((Data_File->File = Locate_File (Data_File->Filename, READ_TXTFILE_STRING,"","",NULL,TRUE)) == NULL)
     {
       f = Data_File->Filename;
       Data_File->Filename = NULL;  /* Keeps from closing failed file. */
       Stage=STAGE_INCLUDE_ERR;
       Error ("Cannot reopen file %s on macro return.\n", f);
     }
  }
#endif

  Data_File->Line_Number = Cond_Stack[CS_Index].Line_No;

  Got_EOF=FALSE;

#ifndef FastMacroPatch
  if (fseek(Data_File->File, Cond_Stack[CS_Index].Pos,0) < 0)
  {
    Error("Unable to file seek in return from macro.\n");
  }
#endif

  /* Always destroy macro locals */
  Destroy_Table(Table_Index--);
}

void Destroy_Macro(POV_MACRO *PMac)
{
  int i;
  if (PMac==NULL)
  {
    return;
  }
  
  POV_FREE(PMac->Macro_Name);
  if (PMac->Macro_Filename!=NULL)
  {
     POV_FREE(PMac->Macro_Filename);
  } 
  
  for (i=0; i < PMac->Num_Of_Pars; i++)
  {
    POV_FREE(PMac->Par_Name[i]);
  }

#ifdef FastMacroPatch
  if(PMac->Text != NULL)
    POV_FREE(PMac->Text);
#endif
  POV_FREE(PMac);
}

POV_ARRAY *Parse_Array_Declare (void)
{
  POV_ARRAY *New;
  int i,j;
  
  New=(POV_ARRAY *)POV_MALLOC(sizeof(POV_ARRAY),"array");
  
  i=0;
  j=1;
  
  Ok_To_Declare = FALSE;

  EXPECT
    CASE (LEFT_SQUARE_TOKEN)
      if (i>4)
      {
         Error("Too many array dimensions");
      }
      New->Sizes[i]=(int)(1.0e-08+Parse_Float());
      j *= New->Sizes[i];
      i++;
      GET(RIGHT_SQUARE_TOKEN)
    END_CASE
    
    OTHERWISE
      UNGET
      EXIT
    END_CASE
  END_EXPECT

  New->Dims     = i-1;
  New->Total    = j;
  New->Type     = EMPTY_ARRAY_TOKEN;
  New->DataPtrs = (void **)POV_MALLOC(sizeof(void *)*j,"array");

  j = 1;
             
  for(i = New->Dims; i>=0; i--)
  {
     New->Mags[i] = j;
     j *= New->Sizes[i];
  }
  
  for (i=0; i<New->Total; i++)
  {
     New->DataPtrs[i] = NULL;
  }
  
  EXPECT
    CASE(LEFT_CURLY_TOKEN)
      UNGET
        Parse_Initalizer(0,0,New);      
      EXIT
    END_CASE
    
    OTHERWISE
      UNGET
      EXIT
    END_CASE
  END_EXPECT
             
  Ok_To_Declare = TRUE;
  return(New);
  
}

static void Parse_Initalizer (int Sub, int Base, POV_ARRAY *a)
{
  int i;

  Parse_Begin();
  if (Sub < a->Dims)
  {
     for(i=0; i < a->Sizes[Sub]; i++)
     {
        Parse_Initalizer(Sub+1,i*a->Mags[Sub]+Base,a);
     }
  }
  else
  {
     for(i=0; i < a->Sizes[Sub]; i++)
     {
        if (!Parse_RValue (a->Type,&(a->Type),&(a->DataPtrs[Base+i]), FALSE,FALSE))
        {
          Error("Insufficent number of initializers");
        }
        Parse_Comma();
     }
  }
  Parse_End();
  Parse_Comma();
}

static void Parse_Fopen(void)
{
   DATA_FILE *New;
   char *temp;
   SYM_ENTRY *Entry;
   
   New=(DATA_FILE *)POV_MALLOC(sizeof(DATA_FILE),"user file");
#ifdef FastMacroPatch
   Fopen_Setup (New);
#else
   New->File=NULL;
#endif
   New->Filename=NULL;

   GET(IDENTIFIER_TOKEN)
   Entry = Add_Symbol (1,Token.Token_String,FILE_ID_TOKEN);
   Entry->Data=(void *)New;

   temp = New->Filename=Parse_String();

   EXPECT
     CASE(READ_TOKEN)
       New->R_Flag = TRUE;
#ifdef FastMacroPatch
       New->Data = Locate_File (temp, READ_TXTFILE_STRING,"","",NULL,TRUE);
#else
       New->File = Locate_File (temp, READ_TXTFILE_STRING,"","",NULL,TRUE);
#endif
       EXIT
     END_CASE
     
     CASE(WRITE_TOKEN)
       New->R_Flag = FALSE;
#ifdef FastMacroPatch
       New->Data = fopen(temp, WRITE_TXTFILE_STRING);
#else
       New->File= fopen(temp, WRITE_TXTFILE_STRING);
#endif
       EXIT
     END_CASE
     
     CASE(APPEND_TOKEN)
       New->R_Flag = FALSE;
#ifdef FastMacroPatch
       New->Data = fopen(temp, APPEND_TXTFILE_STRING);
#else
       New->File= fopen(temp, APPEND_TXTFILE_STRING);
#endif
       EXIT
     END_CASE
     
     OTHERWISE
       Parse_Error_Str("read or write");
     END_CASE
   END_EXPECT

#ifdef FastMacroPatch
   if (New->Data == NULL)
#else
   if (New->File == NULL)
#endif
   {
      New->Filename=NULL;
      Error ("Cannot open user file %s.\n", temp);
   }
   New->Line_Number     = 0;
}

static void Parse_Fclose(void)
{
   DATA_FILE *Data;
   
   EXPECT
     CASE(FILE_ID_TOKEN)
       Data=(DATA_FILE *)Token.Data;
#ifdef FastMacroPatch
       Data_File_Close (Data);
#else
       fflush(Data->File);
       fclose(Data->File);
       Data->File = NULL;       /* <--- this line added -hdf- */
#endif
       Got_EOF=FALSE;
       Remove_Symbol (1,Token.Token_String);
       EXIT
     END_CASE
     
     OTHERWISE
       EXIT
     END_CASE
   END_EXPECT
}

static void Parse_Read()
{
   DATA_FILE *User_File;
   SYM_ENTRY *Temp_Entry;
   int End_File=FALSE;
   char *File_Id;

   GET(LEFT_PAREN_TOKEN)

   GET(FILE_ID_TOKEN)
   User_File=(DATA_FILE *)Token.Data;
   File_Id=POV_STRDUP(Token.Token_String);

   Parse_Comma(); /* Scene file comma between File_Id and 1st data ident */
   
   LValue_Ok = TRUE;
   
   EXPECT
     CASE (IDENTIFIER_TOKEN)
       if (!End_File)
       {
          Temp_Entry = Add_Symbol (1,Token.Token_String,IDENTIFIER_TOKEN);
          End_File=Parse_Read_Value (User_File,Token.Token_Id, &(Temp_Entry->Token_Number), &(Temp_Entry->Data));
          if (Parse_Comma_RParen())  /* Scene file comma between 2 idents */
          {
            EXIT
          }
       }            
     END_CASE

     CASE (STRING_ID_TOKEN)
       if (!End_File)
       {            
          End_File=Parse_Read_Value (User_File,Token.Token_Id,Token.NumberPtr,Token.DataPtr);
          if (Parse_Comma_RParen()) /* Scene file comma between 2 idents */
          {
            EXIT
          }
       }            
     END_CASE

     CASE2 (VECTOR_FUNCT_TOKEN,FLOAT_FUNCT_TOKEN)
       switch(Token.Function_Id)
       {
         case VECTOR_ID_TOKEN:
         case FLOAT_ID_TOKEN:
           if (!End_File)
           {
              End_File=Parse_Read_Value (User_File,Token.Function_Id,Token.NumberPtr,Token.DataPtr);
              if (Parse_Comma_RParen()) /* Scene file comma between 2 idents */
              {
                EXIT
              }
           }            
           break;

         default:
           Parse_Error(IDENTIFIER_TOKEN);
           break;
       }
     END_CASE
     
     CASE(COMMA_TOKEN)
       if (!End_File)
       {            
          Parse_Error(IDENTIFIER_TOKEN);
       }            
     END_CASE

     CASE(RIGHT_PAREN_TOKEN)
       EXIT
     END_CASE

     OTHERWISE
       Parse_Error(IDENTIFIER_TOKEN);
     END_CASE
   END_EXPECT

   LValue_Ok = FALSE;

   if (End_File)
   {
#ifdef FastMacroPatch
      Data_File_Close(User_File);
#else
      fclose(User_File->File);
      User_File->File = NULL;   /* <--- this line added -hdf- */
#endif
      Got_EOF=FALSE;
      Remove_Symbol (1,File_Id);
   }
   POV_FREE(File_Id);
}

static int Parse_Read_Value(DATA_FILE *User_File,int Previous,int *NumberPtr,void **DataPtr)
{
  /* DATA_FILE *Temp;*/ /*YS not used*/
   DBL Val;
   int End_File=FALSE;
   int i;
   EXPRESS Express;
   
   Temp_Data_File_For_Fread = Data_File;
   Data_File = User_File;
   
   EXPECT
     CASE3 (PLUS_TOKEN,DASH_TOKEN,FLOAT_FUNCT_TOKEN)
       UNGET
       Val=Parse_Signed_Float();
       *NumberPtr = FLOAT_ID_TOKEN;
       Test_Redefine(Previous,NumberPtr,*DataPtr);
       *DataPtr   = (void *) Create_Float();
       *((DBL *)*DataPtr) = Val;
       Parse_Comma(); /* data file comma between 2 data items  */
       EXIT
     END_CASE
     
     CASE (LEFT_ANGLE_TOKEN)
       i=1;
       Express[X]=Parse_Signed_Float();  Parse_Comma();
       Express[Y]=Parse_Signed_Float();  Parse_Comma();

       EXPECT
         CASE3 (PLUS_TOKEN,DASH_TOKEN,FLOAT_FUNCT_TOKEN)
           UNGET
           if (++i>4)
           {
              Error("Vector data too long");
           }
           Express[i]=Parse_Signed_Float(); Parse_Comma();
         END_CASE
         
         CASE (RIGHT_ANGLE_TOKEN)
           EXIT
         END_CASE
         
         OTHERWISE
           Parse_Error_Str("vector");
         END_CASE
       END_EXPECT
       
       switch(i)
       {
          case 1:
            *NumberPtr = UV_ID_TOKEN;
            Test_Redefine(Previous,NumberPtr,*DataPtr);
            *DataPtr   = (void *) Create_UV_Vect();
            Assign_UV_Vect(*DataPtr, Express);
            break;

          case 2:
           *NumberPtr = VECTOR_ID_TOKEN;
           Test_Redefine(Previous,NumberPtr,*DataPtr);
           *DataPtr   = (void *) Create_Vector();
            Assign_Vector(*DataPtr, Express);
            break;
            
          case 3:
            *NumberPtr = VECTOR_4D_ID_TOKEN;
            Test_Redefine(Previous,NumberPtr,*DataPtr);
            *DataPtr   = (void *) Create_Vector_4D();
            Assign_Vector_4D(*DataPtr, Express);
            break;
            
          case 4:
            *NumberPtr    = COLOUR_ID_TOKEN;
            Test_Redefine(Previous,NumberPtr,*DataPtr);
            *DataPtr      = (void *) Create_Colour();
            Assign_Colour_Express((COLC*)(*DataPtr), Express); /* NK fix assign_colour bug */
            break;
       }

       Parse_Comma(); /* data file comma between 2 data items  */
       EXIT
     END_CASE

     CASE(STRING_LITERAL_TOKEN)
       *NumberPtr = STRING_ID_TOKEN;
       Test_Redefine(Previous,NumberPtr,*DataPtr);
       *DataPtr   = POV_MALLOC(strlen(Token.Token_String) + 1, "temporary string");
       strcpy ((char *)*DataPtr, Token.Token_String);
       Parse_Comma(); /* data file comma between 2 data items  */
       EXIT
     END_CASE
     
     CASE (END_OF_FILE_TOKEN)
       EXIT
     END_CASE

     OTHERWISE
       Parse_Error_Str ("float, vector, or string literal");
     END_CASE
   END_EXPECT
   
   if (Token.Token_Id==END_OF_FILE_TOKEN)
   {
      End_File = TRUE;
   }
   
   Token.End_Of_File = FALSE;
   Token.Unget_Token = FALSE;
   Got_EOF           = FALSE;

   Data_File = Temp_Data_File_For_Fread;
   Temp_Data_File_For_Fread = NULL;
   
   return(End_File);
}

static int Parse_Comma_RParen(void)
{
  int Stop=FALSE;
  int Old_Ok=Ok_To_Declare;
  
  Ok_To_Declare=FALSE;

  EXPECT
    CASE(COMMA_TOKEN)
      EXIT
    END_CASE
    
    CASE(RIGHT_PAREN_TOKEN)
      Stop=TRUE;
      EXIT
    END_CASE
    
    OTHERWISE
      Parse_Error_Str("comma or right paren");
    END_CASE
  END_EXPECT

  Ok_To_Declare=Old_Ok;
  
  return(Stop);  
}

static void Parse_Write(void)
{
   char *temp;
   DATA_FILE *User_File;
   EXPRESS Express;
   int Terms;

   GET(LEFT_PAREN_TOKEN)
   GET(FILE_ID_TOKEN)
   
   User_File=(DATA_FILE *)Token.Data;

   Parse_Comma();
   
   EXPECT
#ifdef TimeDatePatch
    CASE(GET_DATE_TOKEN) /*YS get date and time string*/
#endif
     CASE4 (STRING_LITERAL_TOKEN,CHR_TOKEN,SUBSTR_TOKEN,STR_TOKEN)
     CASE4 (CONCAT_TOKEN,STRUPR_TOKEN,STRLWR_TOKEN,STRING_ID_TOKEN)
       UNGET
       temp=Parse_Formatted_String();
#ifdef FastMacroPatch
       Data_File_Printf(User_File, temp);
#else
       fprintf(User_File->File,temp);
#endif
       POV_FREE(temp);
       if (Parse_Comma_RParen())
       {
         EXIT
       }
     END_CASE
     
     CASE_VECTOR
       Terms = Parse_Unknown_Vector (Express);
       switch (Terms)
       {
         case 1:
#ifdef FastMacroPatch
           Data_File_Printf(User_File,"%g", Express[X]);
#else
           fprintf(User_File->File,"%g",Express[X]);
#endif
           break;
           
         case 2:
#ifdef FastMacroPatch
           Data_File_Printf(User_File,"<%g,%g> ",Express[U],Express[V]);
#else
           fprintf(User_File->File,"<%g,%g> ",Express[U],Express[V]);
#endif
           break;

         case 3:
#ifdef FastMacroPatch
           Data_File_Printf(User_File,"<%g,%g,%g> ",Express[X],Express[Y],Express[Z]);
#else
           fprintf(User_File->File,"<%g,%g,%g> ",Express[X],Express[Y],Express[Z]);
#endif
           break;

         case 4:
#ifdef FastMacroPatch
           Data_File_Printf(User_File,"<%g,%g,%g,%g> ",Express[X],Express[Y],Express[Z],Express[T]);
#else
           fprintf(User_File->File,"<%g,%g,%g,%g> ",Express[X],Express[Y],Express[Z],Express[T]);
#endif
           break;

         case 5:
#ifdef FastMacroPatch
           Data_File_Printf(User_File,"<%g,%g,%g,%g,%g> ",Express[X],Express[Y],Express[Z],Express[3],Express[4]);
#else
           fprintf(User_File->File,"<%g,%g,%g,%g,%g> ",Express[X],Express[Y],Express[Z],Express[3],Express[4]);
#endif
           break;
         
         default:
           Parse_Error_Str("expression");
       }
       if (Parse_Comma_RParen())
       {
         EXIT
       }
     END_CASE

     OTHERWISE
       Parse_Error_Str("string or expression");
     END_CASE
   END_EXPECT
}

static DBL Parse_Cond_Param(void)
{
  int Old_Ok = Ok_To_Declare;
  int Old_Sk = Skipping;
  DBL Val;
  
/*#ifndef FastMacroPatch*/
  Ok_To_Declare = FALSE;
/*#endif*/
  Skipping      = FALSE;

  Val=Parse_Float_Param();

  Ok_To_Declare = Old_Ok;
  Skipping      = Old_Sk;
  
  return(Val);
}

static void Parse_Cond_Param2(DBL *V1,DBL *V2)
{
  int Old_Ok = Ok_To_Declare;
  int Old_Sk = Skipping;

  Ok_To_Declare = FALSE;
  Skipping      = FALSE;

  Parse_Float_Param2(V1,V2);

  Ok_To_Declare = Old_Ok;
  Skipping      = Old_Sk;
}

static void Inc_CS_Index(void)
{
  if (++CS_Index >= COND_STACK_SIZE)
  {
    Error("Too many nested conditionals or macros.\n");
  }
}

int Parse_Ifdef_Param (void)
{
  int Local_Index;

  SYM_ENTRY *Entry;
  int retval = FALSE;
  int i,j,k;
  register int c;
  DBL val;
  POV_ARRAY *a;

  GET(LEFT_PAREN_TOKEN)
  Inside_Ifdef=TRUE;
  Get_Token();
  strcpy(String2,String);
  Inside_Ifdef=FALSE;
  /*GET(RIGHT_PAREN_TOKEN)*/

  /* Search tables from newest to oldest */
  /*for (Local_Index=Table_Index; Local_Index > 0; Local_Index--)
  {
   if (Find_Symbol(Local_Index,String2) != NULL)
   {
     return(TRUE);
   }
  }

  return(FALSE);
  */
  for (Local_Index=Table_Index; Local_Index > 0; Local_Index--) 
  {
    Entry = Find_Symbol(Local_Index,String2);
    if ( Entry != NULL) 
    {
      Token.Token_Id  =   Entry->Token_Number;
      Token.NumberPtr = &(Entry->Token_Number);
      Token.DataPtr   = &(Entry->Data);

	    if ( Token.Token_Id == PARAMETER_ID_TOKEN ) 
      {
		    Token.NumberPtr = ((POV_PARAM *)(Entry->Data))->NumberPtr;
		    Token.DataPtr = ((POV_PARAM *)(Entry->Data))->DataPtr;
      }
      if (Token.NumberPtr && *(Token.NumberPtr)==ARRAY_ID_TOKEN) 
      {
        Skip_Spaces();
        c = Echo_getc();
        Echo_ungetc(c);

        if (c!='[')
        {
          retval = TRUE; 
	    		break;
        }

        a = (POV_ARRAY *)(*(Token.DataPtr));
        j = 0;
      
        for (i=0; i <= a->Dims; i++)
        {
          GET(LEFT_SQUARE_TOKEN)
          val=Parse_Float();
          if (val<0.0)
          {
            Error("Negative subscript");
          }
          k=(int)(1.0e-08+val);
              
          if (k>=a->Sizes[i])
          {
            Error("Array subscript out of range");
          }
          j += k * a->Mags[i];
          GET(RIGHT_SQUARE_TOKEN)
        }
          
        Token.DataPtr   = &(a->DataPtrs[j]);
        Token.NumberPtr = &(a->Type);
        Token.Token_Id = a->Type;

        retval = (*Token.DataPtr != NULL);
		    break;
      }
      else 
      {
        retval = TRUE;
	    	break;
      }
      
    }
  }

  GET(RIGHT_PAREN_TOKEN)
  return(retval);
}

/* NK persist - Nov 1999 */
int getFilePos(void)
{
#ifdef FastMacroPatch
  return(Data_File_Tell(Data_File));
#else
  return(ftell(Data_File->File));
#endif
}

void seekFilePos(int p)
{
#ifdef FastMacroPatch
  Data_File_Seek(Data_File,p);
#else
  fseek(Data_File->File,p,0);
#endif
}

static int lineNum;
static int filePos;
void saveFilePos()
{
  lineNum = Data_File->Line_Number;
  filePos = 
#ifdef FastMacroPatch
    Data_File_Tell(Data_File);
#else
    ftell(Data_File->File);
#endif
}
void restoreFilePos()
{
#ifdef FastMacroPatch
  Data_File_Seek(Data_File,filePos);
#else
  fseek(Data_File->File,filePos,0);
#endif
  Data_File->Line_Number = lineNum;
}
/* NK ---- */

#ifdef UnofficialBlocking
int getUnofficialVersion(void)
{
  return Data_File->unofficialVersion;
}

void parseUnofficialFeature(int version)
{
  if (Data_File->unofficialVersion < version)
  {
    Error("This feature requires %s %g.\nSpecify '#version unofficial %s %g;' IN THIS FILE to enable this feature.\n",
      UNOFFICIAL_VERSION_STRING,version/100.0,UNOFFICIAL_VERSION_STRING,UNOFFICIAL_VERSION_NUMBER/100.0);
  }
}
#endif


#ifdef FastMacroPatch

/* Close the data file */
static void Fopen_Close (DATA_FILE *File)
{
  if(File->Data!=NULL)
  {
    fflush ((FILE *) File->Data);
    fclose ((FILE *) File->Data);
    File->Data = NULL;
  }
}

/* Read a character */
static int Fopen_Getc (DATA_FILE *File)
{
  int Ch;
  if ((Ch = getc ((FILE *) File->Data)) == EOF)
    return -1;
  return Ch;
}

/* Stuff a character back into the file */
#ifdef __GO32__
static void Fopen_Ungetc (DATA_FILE *File, int Ch)
{
/*
  if(Ch=='\n')
  {
    ungetc ('\r', (FILE *) File->Data);
  }
*/
  ungetc (Ch, (FILE *) File->Data);
}
#else
static void Fopen_Ungetc (DATA_FILE *File, int Ch)
{
  ungetc (Ch, (FILE *) File->Data);
}
#endif

static void Fopen_Putc (DATA_FILE *File, int Ch)
{
  putc (Ch, (FILE *) File->Data);
}

static int Fopen_Tell (DATA_FILE *File)
{
  return ftell ((FILE *) File->Data);
}

static void Fopen_Setup (DATA_FILE *File)
{
  File->Data = NULL;
  File->Tell = Fopen_Tell;
  File->Putc = Fopen_Putc;
  File->Close = Fopen_Close;
  File->Ungetc = Fopen_Ungetc;
  File->Getc = Fopen_Getc;
  File->Seek = Fopen_Seek;
}

static int Fopen_Seek (DATA_FILE *File, int Offset)
{
  return fseek ((FILE *) File->Data, Offset, 0);
}

/* This is the only instance in POV-Ray of it using "..." on it's own */
static void Data_File_Printf (DATA_FILE *File, char *str, ...)
{
  char buf [1024], *ptr = buf;
  va_list arg;
  int len;

  va_start (arg, str);
  len = vsprintf (buf, str, arg);
  va_end (arg);
  while (len --)
    Data_File_Putc (File, *ptr ++);
}

/* Close the data file */
static void String_Close (DATA_FILE *File)
{
  if (File->Data!=NULL)
    POV_FREE (File->Data);
  File->Data = NULL;
}

/* Read a character */
static int String_Getc (DATA_FILE *File)
{
  int Ch = **(char **) File->Data;
  if (Ch == '\0')
    return '\0';
  (*(char **) File->Data) ++;
  return Ch;
}

/* Stuff a character back into the file */
static void String_Ungetc (DATA_FILE *File, int Ch)
{
  (*(char **) File->Data) --;

  /* don't unget too far */
  if  (((char**)File->Data)[0] < ((char**)File->Data)[1])
    ((char**)File->Data)[0] = ((char**)File->Data)[1];
}

static void String_Putc (DATA_FILE *File, int Ch)
{
  /* This cannot occur, so ignore it */
}

static int String_Tell (DATA_FILE *File)
{
  return ((char **) File->Data) [0] - ((char **) File->Data) [1];
}

static int String_Seek (DATA_FILE *File, int Offset)
{
  ((char **) File->Data) [0] = ((char **) File->Data) [1] + Offset;
  return 0;
}

#endif
